from copy import copy
import csv
import os

import pytz

from efootprint.constants.units import u
from efootprint.constants.sources import Source,Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceTimezone
from efootprint.core.country import Country

root_dir = os.path.dirname(os.path.abspath(__file__))
countries_data_file = os.path.join(
    root_dir, "countries_computations", "countries_elec_carbon_intensity_and_timezone.csv")


def tz(timezone: str):
    return SourceTimezone(pytz.timezone(timezone), Sources.USER_DATA)


def country_generator(country_name, country_short_name, country_avg_carbon_int, timezone):
    def return_country():
        return Country(country_name, country_short_name, copy(country_avg_carbon_int), copy(timezone))

    return return_country


def country_generator_from_csv(country_name):
    with open(countries_data_file, newline="", encoding="utf-8") as f:
        countries_data_dict = csv.DictReader(f)
        for row in countries_data_dict:
            if row["Entity"] == country_name:
                country_short_name = row["Code"]
                source = Source(f"Our world in data ({row["Year"]})",
                                "https://ourworldindata.org/grapher/carbon-intensity-electricity")
                country_carbon_int = SourceValue(
                    int(float(row["Carbon intensity of electricity - gCO2/kWh"])) * u.g / u.kWh, source)
                timezone = tz(row["Timezone"])
                break

    return country_generator(country_name, country_short_name, copy(country_carbon_int), copy(timezone))


class Countries:
    # All countries with more than 5 million population in 2024
    AFGHANISTAN = country_generator_from_csv("Afghanistan")
    ALGERIA = country_generator_from_csv("Algeria")
    ANGOLA = country_generator_from_csv("Angola")
    ARGENTINA = country_generator_from_csv("Argentina")
    AUSTRALIA = country_generator_from_csv("Australia")
    AUSTRIA = country_generator_from_csv("Austria")
    AZERBAIJAN = country_generator_from_csv("Azerbaijan")
    BANGLADESH = country_generator_from_csv("Bangladesh")
    BELARUS = country_generator_from_csv("Belarus")
    BELGIUM = country_generator_from_csv("Belgium")
    BENIN = country_generator_from_csv("Benin")
    BOLIVIA = country_generator_from_csv("Bolivia")
    BRAZIL = country_generator_from_csv("Brazil")
    BULGARIA = country_generator_from_csv("Bulgaria")
    BURKINA_FASO = country_generator_from_csv("Burkina Faso")
    BURUNDI = country_generator_from_csv("Burundi")
    CAMBODIA = country_generator_from_csv("Cambodia")
    CAMEROON = country_generator_from_csv("Cameroon")
    CANADA = country_generator_from_csv("Canada")
    CENTRAL_AFRICAN_REPUBLIC = country_generator_from_csv("Central African Republic")
    CHAD = country_generator_from_csv("Chad")
    CHILE = country_generator_from_csv("Chile")
    CHINA = country_generator_from_csv("China")
    COLOMBIA = country_generator_from_csv("Colombia")
    CONGO = country_generator_from_csv("Congo")
    COSTA_RICA = country_generator_from_csv("Costa Rica")
    COTE_D_IVOIRE = country_generator_from_csv("Cote d\'Ivoire")
    CUBA = country_generator_from_csv("Cuba")
    CZECHIA = country_generator_from_csv("Czechia")
    DEMOCRATIC_REPUBLIC_OF_CONGO = country_generator_from_csv("Democratic Republic of Congo")
    DENMARK = country_generator_from_csv("Denmark")
    DOMINICAN_REPUBLIC = country_generator_from_csv("Dominican Republic")
    ECUADOR = country_generator_from_csv("Ecuador")
    EGYPT = country_generator_from_csv("Egypt")
    EL_SALVADOR = country_generator_from_csv("El Salvador")
    ETHIOPIA = country_generator_from_csv("Ethiopia")
    FINLAND = country_generator_from_csv("Finland")
    FRANCE = country_generator_from_csv("France")
    GERMANY = country_generator_from_csv("Germany")
    GHANA = country_generator_from_csv("Ghana")
    GREECE = country_generator_from_csv("Greece")
    GUATEMALA = country_generator_from_csv("Guatemala")
    GUINEA = country_generator_from_csv("Guinea")
    HAITI = country_generator_from_csv("Haiti")
    HONDURAS = country_generator_from_csv("Honduras")
    HONG_KONG = country_generator_from_csv("Hong Kong")
    HUNGARY = country_generator_from_csv("Hungary")
    INDIA = country_generator_from_csv("India")
    INDONESIA = country_generator_from_csv("Indonesia")
    IRAN = country_generator_from_csv("Iran")
    IRAQ = country_generator_from_csv("Iraq")
    IRELAND = country_generator_from_csv("Ireland")
    ISRAEL = country_generator_from_csv("Israel")
    ITALY = country_generator_from_csv("Italy")
    JAPAN = country_generator_from_csv("Japan")
    JORDAN = country_generator_from_csv("Jordan")
    KAZAKHSTAN = country_generator_from_csv("Kazakhstan")
    KENYA = country_generator_from_csv("Kenya")
    KYRGYZSTAN = country_generator_from_csv("Kyrgyzstan")
    LAOS = country_generator_from_csv("Laos")
    LEBANON = country_generator_from_csv("Lebanon")
    LIBERIA = country_generator_from_csv("Liberia")
    LIBYA = country_generator_from_csv("Libya")
    MADAGASCAR = country_generator_from_csv("Madagascar")
    MALAWI = country_generator_from_csv("Malawi")
    MALAYSIA = country_generator_from_csv("Malaysia")
    MALI = country_generator_from_csv("Mali")
    MAURITANIA = country_generator_from_csv("Mauritania")
    MEXICO = country_generator_from_csv("Mexico")
    MOROCCO = country_generator_from_csv("Morocco")
    MOZAMBIQUE = country_generator_from_csv("Mozambique")
    MYANMAR = country_generator_from_csv("Myanmar")
    NEPAL = country_generator_from_csv("Nepal")
    NETHERLANDS = country_generator_from_csv("Netherlands")
    NEW_ZEALAND = country_generator_from_csv("New Zealand")
    NICARAGUA = country_generator_from_csv("Nicaragua")
    NIGER = country_generator_from_csv("Niger")
    NIGERIA = country_generator_from_csv("Nigeria")
    NORTH_KOREA = country_generator_from_csv("North Korea")
    NORWAY = country_generator_from_csv("Norway")
    OMAN = country_generator_from_csv("Oman")
    PAKISTAN = country_generator_from_csv("Pakistan")
    PALESTINE = country_generator_from_csv("Palestine")
    PAPUA_NEW_GUINEA = country_generator_from_csv("Papua New Guinea")
    PARAGUAY = country_generator_from_csv("Paraguay")
    PERU = country_generator_from_csv("Peru")
    PHILIPPINES = country_generator_from_csv("Philippines")
    POLAND = country_generator_from_csv("Poland")
    PORTUGAL = country_generator_from_csv("Portugal")
    ROMANIA = country_generator_from_csv("Romania")
    RUSSIA = country_generator_from_csv("Russia")
    RWANDA = country_generator_from_csv("Rwanda")
    SAUDI_ARABIA = country_generator_from_csv("Saudi Arabia")
    SENEGAL = country_generator_from_csv("Senegal")
    SERBIA = country_generator_from_csv("Serbia")
    SIERRA_LEONE = country_generator_from_csv("Sierra Leone")
    SINGAPORE = country_generator_from_csv("Singapore")
    SLOVAKIA = country_generator_from_csv("Slovakia")
    SOMALIA = country_generator_from_csv("Somalia")
    SOUTH_AFRICA = country_generator_from_csv("South Africa")
    SOUTH_KOREA = country_generator_from_csv("South Korea")
    SOUTH_SUDAN = country_generator_from_csv("South Sudan")
    SPAIN = country_generator_from_csv("Spain")
    SRI_LANKA = country_generator_from_csv("Sri Lanka")
    SUDAN = country_generator_from_csv("Sudan")
    SWEDEN = country_generator_from_csv("Sweden")
    SWITZERLAND = country_generator_from_csv("Switzerland")
    SYRIA = country_generator_from_csv("Syria")
    TAJIKISTAN = country_generator_from_csv("Tajikistan")
    TANZANIA = country_generator_from_csv("Tanzania")
    THAILAND = country_generator_from_csv("Thailand")
    TOGO = country_generator_from_csv("Togo")
    TUNISIA = country_generator_from_csv("Tunisia")
    TURKEY = country_generator_from_csv("Turkey")
    TURKMENISTAN = country_generator_from_csv("Turkmenistan")
    UGANDA = country_generator_from_csv("Uganda")
    UKRAINE = country_generator_from_csv("Ukraine")
    UNITED_ARAB_EMIRATES = country_generator_from_csv("United Arab Emirates")
    UNITED_KINGDOM = country_generator_from_csv("United Kingdom")
    UNITED_STATES = country_generator_from_csv("United States")
    UZBEKISTAN = country_generator_from_csv("Uzbekistan")
    VENEZUELA = country_generator_from_csv("Venezuela")
    VIETNAM = country_generator_from_csv("Vietnam")
    YEMEN = country_generator_from_csv("Yemen")
    ZAMBIA = country_generator_from_csv("Zambia")
    ZIMBABWE = country_generator_from_csv("Zimbabwe")



from functools import lru_cache

import pint
from pint import UnitRegistry

from efootprint.constants.files import CUSTOM_UNITS_PATH

u = UnitRegistry(CUSTOM_UNITS_PATH)
u.default_locale = 'en_EN'
pint.set_application_registry(u)


@lru_cache(maxsize=None)
def get_unit(unit_str):
    return u(unit_str)



from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.edge.edge_component import EdgeComponent
from efootprint.core.hardware.hardware_base import InsufficientCapacityError

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern


class EdgeRAMComponent(EdgeComponent):
    compatible_root_units = [u.bit_ram]
    default_values = {
        "carbon_footprint_fabrication": SourceValue(20 * u.kg),
        "power": SourceValue(10 * u.W),
        "lifespan": SourceValue(6 * u.year),
        "idle_power": SourceValue(2 * u.W),
        "ram": SourceValue(8 * u.GB_ram),
        "base_ram_consumption": SourceValue(1 * u.GB_ram),
    }

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity,
                 ram: ExplainableQuantity, base_ram_consumption: ExplainableQuantity):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, idle_power)
        self.ram = ram.set_label(f"RAM of {self.name}").to(u.GB_ram)
        self.base_ram_consumption = base_ram_consumption.set_label(f"Base RAM consumption of {self.name}")

        self.available_ram_per_instance = EmptyExplainableObject()
        self.unitary_hourly_ram_need_per_usage_pattern = ExplainableObjectDict()

    @property
    def calculated_attributes(self):
        return ["available_ram_per_instance", "unitary_hourly_ram_need_per_usage_pattern"] + super().calculated_attributes

    def update_available_ram_per_instance(self):
        available_ram_per_instance = (self.ram.to(u.GB_ram) - self.base_ram_consumption.to(u.GB_ram))

        if available_ram_per_instance < SourceValue(0 * u.B_ram):
            raise InsufficientCapacityError(self, "RAM", self.ram, self.base_ram_consumption)

        self.available_ram_per_instance = available_ram_per_instance.set_label(
            f"Available RAM per {self.name} instance")

    def update_dict_element_in_unitary_hourly_ram_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_ram_need = sum(
            [need.unitary_hourly_need_per_usage_pattern[usage_pattern]
             for need in self.recurrent_edge_component_needs if usage_pattern in need.edge_usage_patterns],
            start=EmptyExplainableObject())

        if not isinstance(unitary_hourly_ram_need, EmptyExplainableObject):
            max_ram_need = unitary_hourly_ram_need.max().to(u.GB_ram)
            if max_ram_need > self.available_ram_per_instance:
                raise InsufficientCapacityError(self, "RAM", self.available_ram_per_instance, max_ram_need)

        self.unitary_hourly_ram_need_per_usage_pattern[usage_pattern] = unitary_hourly_ram_need.to(u.GB_ram).set_label(
            f"{self.name} hourly RAM need for {usage_pattern.name}").generate_explainable_object_with_logical_dependency(
            self.available_ram_per_instance)

    def update_unitary_hourly_ram_need_per_usage_pattern(self):
        self.unitary_hourly_ram_need_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_ram_need_per_usage_pattern(usage_pattern)

    def update_dict_element_in_unitary_power_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        if usage_pattern in self.unitary_hourly_ram_need_per_usage_pattern:
            ram_need = self.unitary_hourly_ram_need_per_usage_pattern[usage_pattern]
        else:
            ram_need = EmptyExplainableObject()

        if isinstance(ram_need, EmptyExplainableObject):
            unitary_power = self.idle_power
        else:
            ram_workload = (ram_need + self.base_ram_consumption) / self.ram
            unitary_power = self.idle_power + (self.power - self.idle_power) * ram_workload

        self.unitary_power_per_usage_pattern[usage_pattern] = unitary_power.set_label(
            f"{self.name} unitary power for {usage_pattern.name}")

    def update_unitary_power_per_usage_pattern(self):
        self.unitary_power_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_power_per_usage_pattern(usage_pattern)



from copy import copy
from inspect import _empty as empty_annotation, isabstract
from types import UnionType
from typing import List, get_origin, get_args

import efootprint
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict

from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.all_classes_in_order import ALL_EFOOTPRINT_CLASSES
from efootprint.logger import logger
from efootprint.utils.tools import get_init_signature_params


def compute_classes_generation_order(efootprint_classes_dict):
    classes_to_order_dict = copy(efootprint_classes_dict)
    classes_generation_order = []

    while len(classes_to_order_dict) > 0:
        classes_to_append_to_generation_order = []
        for efootprint_class_name, efootprint_class in classes_to_order_dict.items():
            init_sig_params = get_init_signature_params(efootprint_class)
            classes_needed_to_generate_current_class = []
            for init_sig_param_key in init_sig_params:
                annotation = init_sig_params[init_sig_param_key].annotation
                if annotation is empty_annotation or isinstance(annotation, UnionType):
                    continue
                if get_origin(annotation) and get_origin(annotation) in (list, List):
                    param_type = get_args(annotation)[0]
                else:
                    param_type = annotation
                if issubclass(param_type, ModelingObject):
                    if isabstract(param_type):
                        # Case for UsageJourneyStep which has jobs params being abstract (JobBase)
                        for efootprint_class_name_to_check, efootprint_class_to_check in efootprint_classes_dict.items():
                            if issubclass(efootprint_class_to_check, param_type):
                                classes_needed_to_generate_current_class.append(efootprint_class_name_to_check)
                    else:
                        classes_needed_to_generate_current_class.append(param_type.__name__)
            append_to_classes_generation_order = True
            for class_needed in classes_needed_to_generate_current_class:
                if class_needed not in classes_generation_order:
                    append_to_classes_generation_order = False

            if append_to_classes_generation_order:
                classes_to_append_to_generation_order.append(efootprint_class_name)
        for class_to_append in classes_to_append_to_generation_order:
            classes_generation_order.append(class_to_append)
            del classes_to_order_dict[class_to_append]

    return classes_generation_order

def json_to_system(
        system_dict, launch_system_computations=True, efootprint_classes_dict=None):
    if efootprint_classes_dict is None:
        efootprint_classes_dict = {modeling_object_class.__name__: modeling_object_class
                                   for modeling_object_class in ALL_EFOOTPRINT_CLASSES}

    efootprint_version_key = "efootprint_version"
    json_efootprint_version = system_dict.get(efootprint_version_key, None)
    if json_efootprint_version is None:
        logger.warning(
            f"Warning: the JSON file does not contain the key '{efootprint_version_key}'.")
    else:
        json_major_version = int(json_efootprint_version.split(".")[0])
        efootprint_major_version = int(efootprint.__version__.split(".")[0])
        if (json_major_version < efootprint_major_version) and json_major_version >= 9:
            from efootprint.api_utils.version_upgrade_handlers import VERSION_UPGRADE_HANDLERS
            for version in range(json_major_version, efootprint_major_version):
                system_dict = VERSION_UPGRADE_HANDLERS[version](system_dict, efootprint_classes_dict)
        elif json_major_version != efootprint_major_version:
            logger.warning(
                f"Warning: the version of the efootprint library used to generate the JSON file is "
                f"{json_efootprint_version} while the current version of the efootprint library is "
                f"{efootprint.__version__}. Please make sure that the JSON file is compatible with the current version"
                f" of the efootprint library.")

    class_obj_dict = {}
    flat_obj_dict = {}
    explainable_object_dicts_to_create_after_objects_creation = {}

    classes_generation_order = compute_classes_generation_order(efootprint_classes_dict)
    is_loaded_from_system_with_calculated_attributes = False

    for class_key in classes_generation_order:
        if class_key not in system_dict:
            continue
        if class_key not in class_obj_dict:
            class_obj_dict[class_key] = {}
        current_class = efootprint_classes_dict[class_key]
        current_class_dict = {}
        for class_instance_key in system_dict[class_key]:
            new_obj, new_obj_expl_obj_dicts_to_create_after_objects_creation = current_class.from_json_dict(
                system_dict[class_key][class_instance_key], flat_obj_dict, set_trigger_modeling_updates_to_true=False,
                is_loaded_from_system_with_calculated_attributes=is_loaded_from_system_with_calculated_attributes)

            explainable_object_dicts_to_create_after_objects_creation.update(
                new_obj_expl_obj_dicts_to_create_after_objects_creation)

            if not is_loaded_from_system_with_calculated_attributes and len(new_obj.calculated_attributes) > 0:
                if new_obj.calculated_attributes[0] in system_dict[class_key][class_instance_key]:
                    is_loaded_from_system_with_calculated_attributes = True

            if class_key != "System":
                if is_loaded_from_system_with_calculated_attributes:
                    new_obj.trigger_modeling_updates = True
                else:
                    new_obj.after_init()

            current_class_dict[class_instance_key] = new_obj
            flat_obj_dict[class_instance_key] = new_obj

        class_obj_dict[class_key] = current_class_dict

    for (modeling_obj, attr_key), attr_value in explainable_object_dicts_to_create_after_objects_creation.items():
        explainable_object_dict = ExplainableObjectDict(
            {flat_obj_dict[key]: ExplainableObject.from_json_dict(value) for key, value in attr_value.items()})
        modeling_obj.__setattr__(attr_key, explainable_object_dict, check_input_validity=False)
        for explainable_object_item, explainable_object_json \
                in zip(explainable_object_dict.values(), attr_value.values()):
                explainable_object_item.initialize_calculus_graph_data_from_json(explainable_object_json, flat_obj_dict)

    for system in class_obj_dict["System"].values():
        system.set_initial_and_previous_footprints()
        if is_loaded_from_system_with_calculated_attributes:
            system.trigger_modeling_updates = True
        elif launch_system_computations:
            system.after_init()

    return class_obj_dict, flat_obj_dict



from abc import abstractmethod
from typing import List

import numpy as np
from pint import Quantity

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_recurrent_quantities import ExplainableRecurrentQuantities
from efootprint.abstract_modeling_classes.source_objects import SourceRecurrentValues
from efootprint.constants.units import u
from efootprint.builders.hardware.edge.edge_computer import EdgeComputer
from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed
from efootprint.core.usage.edge.recurrent_edge_storage_need import RecurrentEdgeStorageNeed
from efootprint.core.hardware.edge.edge_component import EdgeComponent


class RecurrentEdgeProcessNeed(RecurrentEdgeComponentNeed):
    def __init__(self, name: str, edge_component: EdgeComponent):
        super().__init__(
            name=name,
            edge_component=edge_component,
            recurrent_need=EmptyExplainableObject()
        )

    @property
    def calculated_attributes(self):
        return ["recurrent_need"] + super().calculated_attributes

    @abstractmethod
    def update_recurrent_need(self):
        pass


class RecurrentEdgeProcessRAMNeed(RecurrentEdgeProcessNeed):
    def update_recurrent_need(self):
        if not self.recurrent_edge_device_needs:
            self.recurrent_need = SourceRecurrentValues(Quantity(np.array([0] * 168, dtype=np.float32), u.GB_ram))
            return
        recurrent_edge_device_need = self.recurrent_edge_device_needs[0]
        self.recurrent_need = recurrent_edge_device_need.recurrent_ram_needed.copy().set_label(
            f"{self.name} recurrent need")


class RecurrentEdgeProcessCPUNeed(RecurrentEdgeProcessNeed):
    def update_recurrent_need(self):
        if not self.recurrent_edge_device_needs:
            self.recurrent_need = SourceRecurrentValues(Quantity(np.array([0] * 168, dtype=np.float32), u.cpu_core))
            return
        recurrent_edge_device_need = self.recurrent_edge_device_needs[0]
        self.recurrent_need = recurrent_edge_device_need.recurrent_compute_needed.copy().set_label(
            f"{self.name} recurrent need")


class RecurrentEdgeProcessStorageNeed(RecurrentEdgeProcessNeed, RecurrentEdgeStorageNeed):
    def update_recurrent_need(self):
        if not self.recurrent_edge_device_needs:
            self.recurrent_need = SourceRecurrentValues(Quantity(np.array([0] * 168, dtype=np.float32), u.GB))
            return
        recurrent_edge_device_need = self.recurrent_edge_device_needs[0]
        self.recurrent_need = recurrent_edge_device_need.recurrent_storage_needed.copy().set_label(
            f"{self.name} recurrent need")


class RecurrentEdgeProcess(RecurrentEdgeDeviceNeed):
    default_values = {
        "recurrent_compute_needed": SourceRecurrentValues(Quantity(np.array([1] * 168, dtype=np.float32), u.cpu_core)),
        "recurrent_ram_needed": SourceRecurrentValues(Quantity(np.array([1] * 168, dtype=np.float32), u.GB_ram)),
        "recurrent_storage_needed": SourceRecurrentValues(Quantity(np.array([0] * 168, dtype=np.float32), u.GB)),
    }

    def __init__(self, name: str, edge_device: EdgeComputer,
                 recurrent_compute_needed: ExplainableRecurrentQuantities,
                 recurrent_ram_needed: ExplainableRecurrentQuantities,
                 recurrent_storage_needed: ExplainableRecurrentQuantities):
        super().__init__(
            name=name,
            edge_device=edge_device,
            recurrent_edge_component_needs=[])
        self.recurrent_compute_needed = recurrent_compute_needed.set_label(
            f"Recurrent compute needed for {self.name}")
        self.recurrent_ram_needed = recurrent_ram_needed.set_label(
            f"Recurrent RAM needed for {self.name}")
        self.recurrent_storage_needed = recurrent_storage_needed.set_label(
            f"Recurrent storage needed for {self.name}")

    def after_init(self):
        if not self.recurrent_edge_component_needs:
            ram_need = RecurrentEdgeProcessRAMNeed(
                name=f"{self.name} RAM need", edge_component=self.edge_device.ram_component)
            compute_need = RecurrentEdgeProcessCPUNeed(
                name=f"{self.name} CPU need", edge_component=self.edge_device.cpu_component)
            storage_need = RecurrentEdgeProcessStorageNeed(
                name=f"{self.name} storage need", edge_component=self.edge_device.storage)

            self.recurrent_edge_component_needs = [ram_need, compute_need, storage_need]
        super().after_init()

    @property
    def ram_need(self) -> RecurrentEdgeProcessRAMNeed:
        return next(need for need in self.recurrent_edge_component_needs
                    if isinstance(need, RecurrentEdgeProcessRAMNeed))

    @property
    def compute_need(self) -> RecurrentEdgeProcessCPUNeed:
        return next(need for need in self.recurrent_edge_component_needs
                    if isinstance(need, RecurrentEdgeProcessCPUNeed))

    @property
    def storage_need(self) -> RecurrentEdgeProcessStorageNeed:
        return next(need for need in self.recurrent_edge_component_needs
                    if isinstance(need, RecurrentEdgeProcessStorageNeed))

    @property
    def attribute_update_entanglements(self):
        return {"edge_device": self.generate_component_needs_changes_from_device_change}

    def generate_component_needs_changes_from_device_change(self, change: List[EdgeComputer]):
        old_edge_computer, new_edge_computer = change[0], change[1]
        component_needs_changes = [
            [self.ram_need.edge_component, new_edge_computer.ram_component],
            [self.compute_need.edge_component, new_edge_computer.cpu_component],
            [self.storage_need.edge_component, new_edge_computer.storage],
        ]
        return component_needs_changes

    @property
    def unitary_hourly_compute_need_per_usage_pattern(self):
        return self.compute_need.unitary_hourly_need_per_usage_pattern

    @property
    def unitary_hourly_ram_need_per_usage_pattern(self):
        return self.ram_need.unitary_hourly_need_per_usage_pattern

    @property
    def unitary_hourly_storage_need_per_usage_pattern(self):
        return self.storage_need.unitary_hourly_need_per_usage_pattern



import pytz

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source


@ExplainableObject.register_subclass(lambda d: "zone" in d)
class ExplainableTimezone(ExplainableObject):
    __slots__ = ()

    @classmethod
    def from_json_dict(cls, d):
        source = Source.from_json_dict(d["source"]) if "source" in d else None
        return cls(pytz.timezone(d["zone"]), label=d["label"], source=source)

    def to_json(self, save_calculated_attributes=False):
        output_dict = {"zone": self.value.zone}
        output_dict.update(super().to_json(save_calculated_attributes))

        return output_dict



from abc import abstractmethod
from typing import List, TYPE_CHECKING, Optional

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.hardware.edge.edge_device import EdgeDevice


class EdgeComponent(ModelingObject):
    @classmethod
    @abstractmethod
    def compatible_root_units(self) -> List["str"]:
        """Return list of acceptable pint units for RecurrentEdgeComponentNeed objects linked to this component."""
        pass

    @classmethod
    @abstractmethod
    def default_values(cls):
        pass

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity, power: ExplainableQuantity,
                 lifespan: ExplainableQuantity, idle_power: ExplainableQuantity):
        super().__init__(name)
        self.carbon_footprint_fabrication = carbon_footprint_fabrication.set_label(
            f"Carbon footprint fabrication of {self.name}")
        self.power = power.set_label(f"Power of {self.name}")
        self.lifespan = lifespan.set_label(f"Lifespan of {self.name}")
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.unitary_power_per_usage_pattern = ExplainableObjectDict()
        self.instances_fabrication_footprint_per_usage_pattern = ExplainableObjectDict()
        self.instances_energy_per_usage_pattern = ExplainableObjectDict()
        self.energy_footprint_per_usage_pattern = ExplainableObjectDict()
        self.instances_fabrication_footprint = EmptyExplainableObject()
        self.instances_energy = EmptyExplainableObject()
        self.energy_footprint = EmptyExplainableObject()

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List["EdgeDevice"]:
        if self.edge_device:
            return [self.edge_device]
        return []

    @property
    def calculated_attributes(self):
        return ["unitary_power_per_usage_pattern", "instances_fabrication_footprint_per_usage_pattern",
                "instances_energy_per_usage_pattern", "energy_footprint_per_usage_pattern",
                "instances_fabrication_footprint", "instances_energy", "energy_footprint"]

    @property
    def recurrent_edge_component_needs(self) -> List["RecurrentEdgeComponentNeed"]:
        from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed
        return [container for container in self.modeling_obj_containers
                if isinstance(container, RecurrentEdgeComponentNeed)]

    @property
    def edge_device(self) -> Optional["EdgeDevice"]:
        from efootprint.core.hardware.edge.edge_device import EdgeDevice
        output = None
        for container in self.modeling_obj_containers:
            if isinstance(container, EdgeDevice):
                if output is not None and container != output:
                    raise PermissionError(
                        f"EdgeComponent object can only be associated with one EdgeDevice object but {self.name} "
                        f"is associated "
                        f"with {[mod_obj.name for mod_obj in self.modeling_obj_containers 
                                 if isinstance(mod_obj, EdgeDevice)]}.")
                output = container

        return output

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return list(set(sum([need.edge_usage_patterns for need in self.recurrent_edge_component_needs], start=[])))

    @abstractmethod
    def update_unitary_power_per_usage_pattern(self):
        pass

    def update_dict_element_in_instances_fabrication_footprint_per_usage_pattern(
            self, usage_pattern: "EdgeUsagePattern"):
        component_fabrication_intensity = (self.carbon_footprint_fabrication / self.lifespan)
        nb_instances = usage_pattern.nb_edge_usage_journeys_in_parallel

        instances_fabrication_footprint = (
            nb_instances * component_fabrication_intensity * ExplainableQuantity(1 * u.hour, "one hour"))

        self.instances_fabrication_footprint_per_usage_pattern[usage_pattern] = instances_fabrication_footprint.to(
            u.kg).set_label(f"Hourly {self.name} instances fabrication footprint for {usage_pattern.name}")

    def update_instances_fabrication_footprint_per_usage_pattern(self):
        """Calculate fabrication footprint per usage pattern."""
        self.instances_fabrication_footprint_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_instances_fabrication_footprint_per_usage_pattern(usage_pattern)

    def update_dict_element_in_instances_energy_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        nb_instances = usage_pattern.nb_edge_usage_journeys_in_parallel
        unitary_energy = (self.unitary_power_per_usage_pattern[usage_pattern] *
                        ExplainableQuantity(1 * u.hour, "one hour"))
        instances_energy = nb_instances * unitary_energy

        self.instances_energy_per_usage_pattern[usage_pattern] = instances_energy.set_label(
            f"Hourly energy consumed by {self.name} instances for {usage_pattern.name}")

    def update_instances_energy_per_usage_pattern(self):
        """Calculate energy per usage pattern."""
        self.instances_energy_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_instances_energy_per_usage_pattern(usage_pattern)

    def update_dict_element_in_energy_footprint_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        energy_footprint = (
                self.instances_energy_per_usage_pattern[usage_pattern] * usage_pattern.country.average_carbon_intensity)

        self.energy_footprint_per_usage_pattern[usage_pattern] = energy_footprint.set_label(
            f"{self.name} energy footprint for {usage_pattern.name}")

    def update_energy_footprint_per_usage_pattern(self):
        """Calculate energy footprint per usage pattern."""
        self.energy_footprint_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_energy_footprint_per_usage_pattern(usage_pattern)

    def update_instances_fabrication_footprint(self):
        """Sum fabrication footprint across usage patterns."""
        instances_fabrication_footprint = sum(
            self.instances_fabrication_footprint_per_usage_pattern.values(), start=EmptyExplainableObject())
        self.instances_fabrication_footprint = instances_fabrication_footprint.set_label(
            f"{self.name} total fabrication footprint across usage patterns")

    def update_instances_energy(self):
        """Sum energy across usage patterns."""
        instances_energy = sum(
            self.instances_energy_per_usage_pattern.values(), start=EmptyExplainableObject())
        self.instances_energy = instances_energy.set_label(
            f"{self.name} total energy consumed across usage patterns")

    def update_energy_footprint(self):
        """Sum energy footprint across usage patterns."""
        energy_footprint = sum(
            self.energy_footprint_per_usage_pattern.values(), start=EmptyExplainableObject())
        self.energy_footprint = energy_footprint.set_label(
            f"{self.name} total energy footprint across usage patterns")



from typing import TYPE_CHECKING, List

from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.core.hardware.edge.edge_device import EdgeDevice
from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.usage.edge.edge_function import EdgeFunction


class RecurrentEdgeDeviceNeed(ModelingObject):
    default_values = {}

    def __init__(self, name: str, edge_device: EdgeDevice,
                 recurrent_edge_component_needs: List[RecurrentEdgeComponentNeed]):
        super().__init__(name)
        self.edge_device = edge_device
        self.recurrent_edge_component_needs = recurrent_edge_component_needs

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[RecurrentEdgeComponentNeed]:
        # edge_device is automatically included through components' modeling_objects_whose_attributes_depend_directly_on_me
        return self.recurrent_edge_component_needs

    @property
    def edge_functions(self) -> List["EdgeFunction"]:
        return self.modeling_obj_containers

    @property
    def edge_usage_journeys(self) -> List["EdgeUsageJourney"]:
        return list(set(sum([ef.edge_usage_journeys for ef in self.edge_functions], start=[])))

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return list(set(sum([euj.edge_usage_patterns for euj in self.edge_usage_journeys], start=[])))



from datetime import datetime, timedelta
from typing import List

import numpy as np
from pint import Quantity, Unit
import ciso8601

from efootprint.constants.units import u
from efootprint.abstract_modeling_classes.source_objects import SourceHourlyValues


def create_random_source_hourly_values(
        timespan: Quantity = 1 * u.day, min_val: int = 1, max_val: int = 10,
        start_date: datetime = ciso8601.parse_datetime("2025-01-01"),
        pint_unit: Unit = u.occurrence):
    nb_days = timespan.to(u.day).magnitude
    data = Quantity(
        np.random.randint(min_val, max_val, size=int(nb_days * 24)).astype(np.float32, copy=False), pint_unit)
    shv = SourceHourlyValues(data, start_date=start_date)

    return shv


def create_source_hourly_values_from_list(
        input_list: List[float], start_date: datetime = ciso8601.parse_datetime("2025-01-01"),
        pint_unit: Unit = u.occurrence):
    return SourceHourlyValues(Quantity(np.array(input_list, dtype=np.float32), pint_unit), start_date)


def linear_growth_hourly_values(
        timespan: Quantity, start_value: int, end_value: int,
        start_date: datetime = ciso8601.parse_datetime("2025-01-01"),
        pint_unit: Unit = u.occurrence):
    nb_of_hours = int(timespan.to(u.hour).magnitude)
    linear_growth = np.linspace(start_value, end_value, nb_of_hours, dtype=np.float32)

    return create_source_hourly_values_from_list(linear_growth, start_date, pint_unit)


def sinusoidal_fluct_hourly_values(
        timespan: Quantity, sin_fluct_amplitude: int, sin_fluct_period_in_hours: int,
        start_date: datetime = ciso8601.parse_datetime("2025-01-01"),
        pint_unit: Unit = u.occurrence):
    nb_of_hours = int(timespan.to(u.hour).magnitude)
    time = np.arange(nb_of_hours, dtype=np.float32)
    sinusoidal_fluctuation = (
            sin_fluct_amplitude * np.sin(2 * np.pi * time / sin_fluct_period_in_hours, dtype=np.float32))

    return create_source_hourly_values_from_list(sinusoidal_fluctuation, start_date, pint_unit)


def daily_fluct_hourly_values(
        timespan: Quantity, fluct_scale: float, hour_of_day_for_min_value: int = 4,
        start_date: datetime = ciso8601.parse_datetime("2025-01-01"),
        pint_unit: Unit = u.occurrence):
    assert fluct_scale > 0
    assert fluct_scale <= 1
    nb_of_hours = int(timespan.to(u.hour).magnitude)
    time = np.arange(nb_of_hours, dtype=np.float32)
    hour_of_day = [(start_date.hour + x) % 24 for x in time]

    daily_fluctuation = (
            np.full(shape=len(hour_of_day), fill_value=1.0, dtype=np.float32)
            + fluct_scale * np.sin(
                (3 * np.pi / 2)
                + (2 * np.pi
                    * (hour_of_day - np.full(
                            shape=len(hour_of_day), fill_value=float(hour_of_day_for_min_value), dtype=np.float32))
                    / 24
                ), dtype=np.float32
            )
        )

    return create_source_hourly_values_from_list(daily_fluctuation, start_date, pint_unit)


def create_hourly_usage_from_frequency(
        timespan: Quantity, input_volume: float, frequency: str, active_days: list = None,
        hours: list = None, start_date: datetime = ciso8601.parse_datetime("2025-01-01"),
        pint_unit: Unit = u.occurrence):
    if frequency not in ['daily', 'weekly', 'monthly', 'yearly']:
        raise ValueError(f"frequency must be one of 'daily', 'weekly', 'monthly', or 'yearly', got {frequency}.")

    if frequency == 'daily' and active_days is not None:
        raise ValueError(f"active_days must be None for daily frequency, got {active_days}.")

    if active_days is None:
        if frequency == "weekly":
            active_days = [0]  # default to midnight or Monday
        else:
            active_days = [1]  # default to first day of month or first day of year

    if hours is None:
        hours = [0]  # default to midnight

    period_index = [start_date + timedelta(hours=i) for i in range(int(timespan.to(u.hour).magnitude) + 1)]
    # Important to have fill_value be 0.0 otherwise values will be cast to int
    values = np.full(shape=len(period_index), fill_value=0.0, dtype=np.float32)

    for i, period in enumerate(period_index):
        hour_of_day = period.hour  # Hour of the day, 0 to 23
        day_of_week = period.weekday()  # Day of the week, 0 to 6
        day_of_month = period.day  # Day of the month, 1 to 31
        day_of_year = period.timetuple().tm_yday  # Day of the year, 1 to 365/366
        if frequency == 'daily':
            if hour_of_day in hours:
                values[i] = input_volume
        elif frequency == 'weekly':
            if day_of_week in active_days and hour_of_day in hours:
                values[i] = input_volume
        elif frequency == 'monthly':
            if day_of_month in active_days and hour_of_day in hours:
                values[i] = input_volume
        elif frequency == 'yearly':
            if day_of_year in active_days and hour_of_day in hours:
                values[i] = input_volume

    return SourceHourlyValues(Quantity(values, pint_unit), start_date, label="Hourly usage")


def create_hourly_usage_from_daily_volume_and_list_of_hours(
        timespan: Quantity, daily_volume: float, hours: List[int],
        start_date: datetime = ciso8601.parse_datetime("2025-01-01"),
        pint_unit: Unit = u.occurrence):
    volume_per_hour = daily_volume / len(hours)

    return create_hourly_usage_from_frequency(
        timespan, volume_per_hour, frequency='daily', active_days=None, hours=hours,
        start_date=start_date, pint_unit=pint_unit)



import math
import numbers
import base64
from copy import copy, deepcopy
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import pytz
from pint import Unit, Quantity
import numpy as np
import zstandard as zstd
import ciso8601

from efootprint.abstract_modeling_classes.explainable_object_base_class import (
    ExplainableObject, Source)
from efootprint.abstract_modeling_classes.explainable_timezone import ExplainableTimezone
from efootprint.constants.units import u, get_unit
from efootprint.logger import logger
from efootprint.utils.plot_baseline_and_simulation_data import plot_baseline_and_simulation_data, prepare_data
from efootprint.abstract_modeling_classes.aggregation_utils import validate_timeseries_unit

if TYPE_CHECKING:
    from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
    from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject


def align_temporally_quantity_arrays(
        first_quantity: Quantity, first_start_date: datetime, second_quantity: Quantity, second_start_date: datetime,
        equalize_units: bool = True):
    if equalize_units and first_quantity.units != second_quantity.units:
        second_quantity = second_quantity.to(first_quantity.units)

    first_quantity_array = first_quantity.magnitude.astype(np.float32)
    second_quantity_array = second_quantity.magnitude.astype(np.float32)

    # Align by start_date
    end1 = first_start_date + timedelta(hours=len(first_quantity_array))
    end2 = second_start_date + timedelta(hours=len(second_quantity_array))

    # Compute common start and end
    common_start = min(first_start_date, second_start_date)
    common_end = max(end1, end2)
    total_len = int((common_end - common_start).total_seconds() // 3600)

    # Create aligned zero arrays
    aligned_first_array = np.zeros(total_len, dtype=np.float32)
    aligned_second_array = np.zeros(total_len, dtype=np.float32)

    # Compute insertion positions
    offset_first_quantity = int((first_start_date - common_start).total_seconds() // 3600)
    offset_second_quantity = int((second_start_date - common_start).total_seconds() // 3600)

    aligned_first_array[offset_first_quantity:offset_first_quantity + len(first_quantity_array)] = first_quantity_array
    aligned_second_array[offset_second_quantity:offset_second_quantity + len(second_quantity_array)] = (
        second_quantity_array)

    return aligned_first_array, aligned_second_array, common_start


@ExplainableObject.register_subclass(lambda d: ("values" in d or "compressed_values" in d) and "unit" in d)
class ExplainableHourlyQuantities(ExplainableObject):
    __slots__ = (
        '_ExplainableQuantity',
        '_EmptyExplainableObject',
        'start_date',
        'json_compressed_value_data',
    )

    @classmethod
    def from_json_dict(cls, d):
        source = Source.from_json_dict(d.get("source")) if d.get("source") else None
        if "values" in d:
            value = Quantity(np.array(d["values"], dtype=np.float32), get_unit(d["unit"]))
        elif "compressed_values" in d:
            # start_date and timezone are included to facilitate json dumping if object doesn’t rehydrate
            value = {k: d[k] for k in ["compressed_values", "unit", "start_date", "timezone"]}
        else:
            raise ValueError("Invalid hourly quantity format")
        start_date = ciso8601.parse_datetime(d["start_date"])
        if d.get("timezone", None) is not None:
            start_date = pytz.timezone(d["timezone"]).localize(start_date)

        return cls(value, start_date=start_date, label=d["label"], source=source)

    def __init__(
            self, value: Quantity | dict, start_date: datetime, label: str = None,
            left_parent: ExplainableObject = None, right_parent: ExplainableObject = None, operator: str = None,
            source: Source = None):
        from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
        from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
        self._ExplainableQuantity = ExplainableQuantity
        self._EmptyExplainableObject = EmptyExplainableObject
        self.start_date = start_date
        self.json_compressed_value_data = None
        if isinstance(value, Quantity):
            validate_timeseries_unit(value, label)
            if value.magnitude.dtype != np.float32:
                logger.info(
                    f"converting value {label} to float32. This is surprising, a casting to np.float32 is probably "
                    f"missing somewhere.")
                value = value.magnitude.astype(np.float32, copy=False) * value.units
            super().__init__(value, label, left_parent, right_parent, operator, source)
        elif isinstance(value, dict):
            self.json_compressed_value_data = value
            super().__init__(None, label, left_parent, right_parent, operator, source)
        else:
            raise ValueError(
                f"ExplainableHourlyQuantities values must be Pint Quantities of numpy arrays or dict, got {type(value)}"
            )

    @property
    def value(self):
        if self._value is None and self.json_compressed_value_data is not None:
            decompressed_values = self.decompress_values(self.json_compressed_value_data["compressed_values"])
            self._value = Quantity(decompressed_values, get_unit(self.json_compressed_value_data["unit"]))

        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value
        self.json_compressed_value_data = None

    @value.deleter
    def value(self):
        self._value = None
        self.json_compressed_value_data = None

    def to(self, unit_to_convert_to: Unit):
        self.value = self.value.to(unit_to_convert_to)
        validate_timeseries_unit(self.value, self.label)

        return self

    def generate_explainable_object_with_logical_dependency(self, explainable_condition: "ExplainableObject"):
        return self.__class__(
            value=self.value, start_date=self.start_date, label=self.label, left_parent=self,
            right_parent=explainable_condition, operator="logically dependent on")

    def __round__(self, round_level):
        return ExplainableHourlyQuantities(
            np.round(self.value, round_level).astype(np.float32, copy=False), start_date=self.start_date, label=self.label,
            left_parent=self, operator=f"rounded to {round_level} decimals", source=self.source
        )

    def round(self, round_level):
        self.value = np.round(self.value, round_level).astype(np.float32, copy=False)

        return self

    def return_shifted_hourly_quantities(self, shift_duration: "ExplainableQuantity"):
        shift_hours = math.floor(shift_duration.to(u.hour).magnitude)

        return ExplainableHourlyQuantities(
            copy(self.value),  # Use copy to avoid modifying the original value
            start_date=self.start_date + timedelta(hours=shift_hours),
            label=f"{self.label} shifted by {shift_hours}h" if self.label else None,
            left_parent=self,
            right_parent=shift_duration,
            operator="shifted by"
        )

    @property
    def unit(self):
        return self.value.units

    @property
    def magnitude(self):
        return self.value.magnitude

    @property
    def plot_aggregation_strategy(self) -> str:
        """Determine how to aggregate hourly values into daily values for plotting."""
        from efootprint.abstract_modeling_classes.aggregation_utils import get_plot_aggregation_strategy
        return get_plot_aggregation_strategy(self.unit)

    @property
    def value_as_float_list(self):
        return self.magnitude.tolist()

    def convert_to_utc(self, local_timezone: ExplainableTimezone):
        if self.start_date.tzinfo is None:
            localized_dt = local_timezone.value.localize(self.start_date)
        else:
            localized_dt = self.start_date

        return ExplainableHourlyQuantities(
            self.value, start_date=localized_dt.astimezone(pytz.utc),
            left_parent=self, right_parent=local_timezone, operator="converted to UTC from")

    def sum(self):
        return self._ExplainableQuantity(np.sum(self.value, dtype=np.float32), left_parent=self, operator="sum")

    def mean(self):
        return self._ExplainableQuantity(np.mean(self.value, dtype=np.float32), left_parent=self, operator="mean")

    def max(self):
        return self._ExplainableQuantity(np.max(self.value), left_parent=self, operator="max")

    def min(self):
        return self._ExplainableQuantity(np.min(self.value), left_parent=self, operator="min")

    def abs(self):
        return ExplainableHourlyQuantities(
            np.abs(self.value), start_date=self.start_date, left_parent=self, operator="abs")

    def ceil(self):
        return ExplainableHourlyQuantities(
            np.ceil(self.value), start_date=self.start_date, left_parent=self, operator="ceil")

    def __neg__(self):
        return ExplainableHourlyQuantities(-self.value, start_date=self.start_date, left_parent=self, operator="negate")

    def np_compared_with(self, compared_object, comparator):
        if isinstance(compared_object, self._EmptyExplainableObject):
            compared_values = np.full(len(self.value), fill_value=np.float32(0))
        elif isinstance(compared_object, ExplainableHourlyQuantities):
            assert compared_object.unit == self.unit, f"{compared_object.unit} != {self.unit}"
            compared_values = compared_object.value
            assert self.start_date == compared_object.start_date, \
                f"Cannot compare ExplainableHourlyQuantities with different start dates: " \
                f"{self.start_date} and {compared_object.start_date}"
        else:
            raise ValueError(f"Can only compare ExplainableHourlyQuantities with ExplainableHourlyQuantities or "
                             f"EmptyExplainableObjects, not {type(compared_object)}")

        if comparator == "max":
            result_comparison_np = np.maximum(self.value, compared_values)
        elif comparator == "min":
            result_comparison_np = np.minimum(self.value, compared_values)
        else:
            raise ValueError(f"Comparator {comparator} not implemented in np_compared_with method")

        return ExplainableHourlyQuantities(
            Quantity(result_comparison_np, self.unit),
            start_date=self.start_date,
            label=f"{self.label} compared with {compared_object.label}",
            left_parent=self,
            right_parent=compared_object,
            operator=f"{comparator} compared with"
        )

    def __copy__(self):
        return ExplainableHourlyQuantities(
            self.value.copy(), copy(self.start_date), label=copy(self.label), source=copy(self.source))

    def copy(self):
        return ExplainableHourlyQuantities(
            self.value.copy(), copy(self.start_date), label=self.label, left_parent=self, operator="duplicate")

    def __eq__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            return False
        elif isinstance(other, self._EmptyExplainableObject):
            return False
        if isinstance(other, ExplainableHourlyQuantities):
            aligned_first_array, aligned_second_array, common_start = align_temporally_quantity_arrays(
                self.value, self.start_date, other.value, other.start_date)

            return np.allclose(aligned_first_array, aligned_second_array, atol=10**-3, rtol=10**-6)

        return False

    def __len__(self):
        return len(self.value)

    def __add__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            return ExplainableHourlyQuantities(
                self.value, start_date=self.start_date, label=self.label,
                left_parent=self, operator=""
            )
        elif isinstance(other, self._EmptyExplainableObject):
            return ExplainableHourlyQuantities(
                self.value, start_date=self.start_date, label=self.label,
                left_parent=self, right_parent=other, operator="+"
            )
        elif isinstance(other, ExplainableHourlyQuantities):
            aligned_self, aligned_other, common_start = align_temporally_quantity_arrays(
                self.value, self.start_date, other.value, other.start_date)
            result_array = aligned_self + aligned_other

            return ExplainableHourlyQuantities(
                Quantity(result_array, self.unit), start_date=common_start, label=None,
                left_parent=self, right_parent=other, operator="+")
        elif isinstance(other, self._ExplainableQuantity):
            return ExplainableHourlyQuantities(
                self.value + other.value, start_date=self.start_date, label=None,
                left_parent=self, right_parent=other, operator="+")
        else:
            raise ValueError(f"Can only add another ExplainableHourlyQuantities or scalar 0 or ExplainableQuantity, "
                             f"not {type(other)}")

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            return ExplainableHourlyQuantities(
                self.value, start_date=self.start_date, label=self.label,
                left_parent=self, operator=""
            )
        elif isinstance(other, self._EmptyExplainableObject):
            return ExplainableHourlyQuantities(
                self.value, start_date=self.start_date, label=self.label,
                left_parent=self, right_parent=other, operator="-"
            )
        elif isinstance(other, ExplainableHourlyQuantities):
            aligned_self, aligned_other, common_start = align_temporally_quantity_arrays(
                self.value, self.start_date, other.value, other.start_date)
            result_array = aligned_self - aligned_other

            return ExplainableHourlyQuantities(
                Quantity(result_array, self.unit), start_date=common_start, label=None,
                left_parent=self, right_parent=other, operator="-")
        elif isinstance(other, self._ExplainableQuantity):
            return ExplainableHourlyQuantities(
                self.value - other.value, start_date=self.start_date, label=None,
                left_parent=self, right_parent=other, operator="-")
        else:
            raise ValueError(f"Can only subtract another ExplainableHourlyQuantities or scalar 0 or ExplainableQuantity,"
                             f" not {type(other)}")

    def __rsub__(self, other):
        if isinstance(other, ExplainableHourlyQuantities):
            return other.__sub__(self)
        elif isinstance(other, self._ExplainableQuantity):
            return ExplainableHourlyQuantities(
                other.value - self.value, start_date=self.start_date, label=None,
                left_parent=other, right_parent=self, operator="-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, "
                             f"not with {type(other)}")

    def __mul__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            return 0
        elif isinstance(other, self._EmptyExplainableObject):
            return self._EmptyExplainableObject(left_parent=self, right_parent=other, operator="*")
        elif isinstance(other, self._ExplainableQuantity):
            other_magnitude_to_multiply = other.magnitude
            if not isinstance(other_magnitude_to_multiply, np.float32):
                other_magnitude_to_multiply = np.float32(other_magnitude_to_multiply)
            result_magnitude = self.value.magnitude * other_magnitude_to_multiply
            result_quantity = Quantity(result_magnitude, self.unit * other.value.units)
            return ExplainableHourlyQuantities(
                result_quantity, self.start_date, "", self, other, "*")
        elif isinstance(other, ExplainableHourlyQuantities):
            aligned_self, aligned_other, common_start = align_temporally_quantity_arrays(
                self.value, self.start_date, other.value, other.start_date, equalize_units=False)
            result_array = aligned_self * aligned_other

            return ExplainableHourlyQuantities(
                Quantity(result_array, self.unit * other.unit), start_date=common_start, label=None,
                left_parent=self, right_parent=other, operator="*")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, "
                f"not with {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, ExplainableHourlyQuantities):
            raise NotImplementedError
        elif isinstance(other, self._ExplainableQuantity):
            other_value_to_divide = other.value
            if not isinstance(other_value_to_divide.magnitude, np.float32):
                other_value_to_divide = np.float32(other_value_to_divide.magnitude) * other_value_to_divide.units
            return ExplainableHourlyQuantities(self.value / other_value_to_divide, self.start_date,"", self, other, "/")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, "
                f"not with {type(other)}")

    def __rtruediv__(self, other):
        if isinstance(other, ExplainableHourlyQuantities):
            raise NotImplementedError
        elif isinstance(other, self._ExplainableQuantity):
            other_value_to_divide = other.value
            if not isinstance(other_value_to_divide.magnitude, np.float32):
                other_value_to_divide = np.float32(other_value_to_divide.magnitude) * other_value_to_divide.units
            return ExplainableHourlyQuantities(other_value_to_divide / self.value, self.start_date,"", other, self, "/")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity,"
                f" not with {type(other)}")

    @staticmethod
    def compress_values(values: np.ndarray) -> str:
        if values.dtype != np.float32:
            values = values.astype(np.float32, copy=False)
        cctx = zstd.ZstdCompressor(level=0)
        compressed = cctx.compress(values.tobytes())
        return base64.b64encode(compressed).decode("utf-8")

    @staticmethod
    def decompress_values(compressed_str: str) -> np.ndarray:
        compressed = base64.b64decode(compressed_str)
        dctx = zstd.ZstdDecompressor()
        decompressed = dctx.decompress(compressed)
        return np.frombuffer(decompressed, dtype=np.float32)

    def to_json(self, save_calculated_attributes=False):
        if self.json_compressed_value_data is not None:
            output_dict = deepcopy(self.json_compressed_value_data)
        else:
            output_dict = {
                    "compressed_values": self.compress_values(self.magnitude),
                    "unit": str(self.unit),
                    "start_date": self.start_date.strftime("%Y-%m-%d %H:%M:%S"),
                    "timezone": str(self.start_date.tzinfo) if self.start_date.tzinfo is not None else None,
                }

        output_dict.update(super().to_json(save_calculated_attributes))

        return output_dict

    def __repr__(self):
        return str(self)

    def __str__(self):
        def _round_series_values(input_series: np.array):
            return [str(round(hourly_value.magnitude, 2)) for hourly_value in input_series]

        compact_unit = "{:~}".format(self.unit)
        nb_of_values = len(self.value)
        if nb_of_values < 30:
            rounded_values = _round_series_values(self.value)
            str_rounded_values = "[" + ", ".join(rounded_values) + "]"
        else:
            first_vals = _round_series_values(self.value[:10])
            last_vals = _round_series_values(self.value[-10:])
            str_rounded_values = "first 10 vals [" + ", ".join(first_vals) \
                                 + "],\n    last 10 vals [" + ", ".join(last_vals) + "]"

        return f"{nb_of_values} values from {self.start_date} " \
               f"to {self.start_date + timedelta(hours=len(self.value))} in {compact_unit}:\n    {str_rounded_values}"

    def plot(self, figsize=(10, 4), filepath=None, plt_show=False, xlims=None, cumsum=False):
        import matplotlib.pyplot as plt

        if self.baseline_twin is None and self.simulation_twin is None:
            baseline_q = self.value
            baseline_start = self.start_date
            simulated_q = None
        elif self.baseline_twin is not None and self.simulation_twin is None:
            baseline_q = self.baseline_twin.value
            baseline_start = self.baseline_twin.start_date
            simulated_q = self.value
            simulated_start = self.start_date
        elif self.simulation_twin is not None and self.baseline_twin is None:
            baseline_q = self.value
            baseline_start = self.start_date
            simulated_q = self.simulation_twin.value
            simulated_start = self.simulation_twin.start_date
        else:
            raise ValueError("Both baseline and simulation twins are not None, this should not happen")

        time_baseline, baseline_q = prepare_data(baseline_q, baseline_start, cumsum)

        if simulated_q is not None:
            if isinstance(simulated_q, self._EmptyExplainableObject):
                sim_len = len(baseline_q.magnitude)
                simulated_q = Quantity(np.zeros(sim_len), baseline_q.units)
                simulated_start = self.simulation.simulation_date

            time_sim, simulated_q = prepare_data(simulated_q, simulated_start, cumsum)

            # Align simulation start offset with baseline in cumsum mode
            if cumsum:
                baseline_interp_index = np.where(time_baseline == time_sim[0])[0]
                if baseline_interp_index.size > 0:
                    simulated_q += baseline_q[baseline_interp_index[0]]
        else:
            time_sim = None

        # Plotting
        fig, ax = plot_baseline_and_simulation_data(baseline_q, time_baseline, simulated_q, time_sim, figsize, xlims)

        if self.label:
            title = self.label if not cumsum else "Cumulative " + self.label[0].lower() + self.label[1:]
            ax.set_title(title)
        if xlims is not None:
            ax.set_xlim(xlims)

        fig.autofmt_xdate()
        if filepath:
            plt.savefig(filepath, bbox_inches='tight')
        if plt_show:
            plt.show()

        return fig, ax



import json

import efootprint
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject


def recursively_write_json_dict(output_dict, mod_obj, save_calculated_attributes):
    mod_obj_class = mod_obj.class_as_simple_str
    if mod_obj_class not in output_dict:
        output_dict[mod_obj_class] = {}
    if mod_obj.id not in output_dict[mod_obj_class]:
        output_dict[mod_obj_class][mod_obj.id] = mod_obj.to_json(save_calculated_attributes)
        for key, value in mod_obj.__dict__.items():
            if isinstance(value, ModelingObject):
                recursively_write_json_dict(output_dict, value, save_calculated_attributes)
            elif isinstance(value, list) and len(value) > 0 and isinstance(value[0], ModelingObject):
                for mod_obj_elt in value:
                    recursively_write_json_dict(output_dict, mod_obj_elt, save_calculated_attributes)

    return output_dict


def system_to_json(input_system, save_calculated_attributes, output_filepath=None, indent=4):
    output_dict = {"efootprint_version": efootprint.__version__}
    recursively_write_json_dict(output_dict, input_system, save_calculated_attributes)

    if output_filepath is not None:
        with open(output_filepath, "w") as file:
            file.write(json.dumps(output_dict, indent=indent))

    return output_dict



from copy import copy

from efootprint.builders.hardware.boavizta_cloud_server import BoaviztaCloudServer
from efootprint.builders.services.service_base_class import Service
from efootprint.core.hardware.edge.edge_device import EdgeDevice
from efootprint.core.hardware.edge.edge_component import EdgeComponent
from efootprint.core.hardware.edge.edge_ram_component import EdgeRAMComponent
from efootprint.core.hardware.edge.edge_cpu_component import EdgeCPUComponent
from efootprint.core.hardware.edge.edge_workload_component import EdgeWorkloadComponent
from efootprint.builders.hardware.edge.edge_appliance import EdgeAppliance, EdgeApplianceComponent
from efootprint.builders.hardware.edge.edge_computer import EdgeComputer, EdgeComputerRAMComponent, EdgeComputerCPUComponent
from efootprint.core.hardware.edge.edge_storage import EdgeStorage
from efootprint.core.hardware.server_base import ServerBase
from efootprint.core.usage.edge.edge_function import EdgeFunction
from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed
from efootprint.core.usage.edge.recurrent_edge_storage_need import RecurrentEdgeStorageNeed
from efootprint.builders.usage.edge.recurrent_edge_process import (
    RecurrentEdgeProcess, RecurrentEdgeProcessRAMNeed, RecurrentEdgeProcessCPUNeed,
    RecurrentEdgeProcessStorageNeed)
from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney
from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed
from efootprint.builders.usage.edge.recurrent_edge_workload import RecurrentEdgeWorkload, RecurrentEdgeWorkloadNeed
from efootprint.core.usage.usage_journey_step import UsageJourneyStep
from efootprint.core.usage.usage_journey import UsageJourney
from efootprint.core.hardware.device import Device
from efootprint.core.country import Country
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.core.hardware.storage import Storage
from efootprint.core.hardware.gpu_server import GPUServer
from efootprint.core.hardware.server import Server
from efootprint.builders.services.generative_ai_ecologits import GenAIModel, GenAIJob
from efootprint.builders.services.video_streaming import VideoStreaming, VideoStreamingJob
from efootprint.builders.services.web_application import WebApplication, WebApplicationJob
from efootprint.core.usage.job import Job, JobBase, GPUJob
from efootprint.core.hardware.network import Network
from efootprint.core.system import System


SERVICE_CLASSES = [WebApplication, VideoStreaming, GenAIModel]
SERVICE_JOB_CLASSES = [WebApplicationJob, VideoStreamingJob, GenAIJob]
SERVER_CLASSES = [Server, GPUServer]
SERVER_BUILDER_CLASSES = [BoaviztaCloudServer]
EDGE_COMPONENT_CLASSES = [EdgeRAMComponent, EdgeCPUComponent, EdgeWorkloadComponent, EdgeStorage]
EDGE_COMPUTER_COMPONENT_CLASSES = [EdgeComputerRAMComponent, EdgeComputerCPUComponent, EdgeApplianceComponent]
RECURRENT_EDGE_COMPONENT_NEED_CLASSES = [
    RecurrentEdgeComponentNeed, RecurrentEdgeProcessRAMNeed, RecurrentEdgeProcessCPUNeed,
    RecurrentEdgeProcessStorageNeed, RecurrentEdgeWorkloadNeed]
RECURRENT_EDGE_DEVICE_NEED_BUILDER_CLASSES = [RecurrentEdgeProcess, RecurrentEdgeWorkload]
EDGE_DEVICE_BUILDER_CLASSES = [EdgeAppliance, EdgeComputer]


ALL_EFOOTPRINT_CLASSES = (
        [UsageJourneyStep, UsageJourney, Device, Country, UsagePattern]
        + [EdgeUsageJourney, EdgeFunction, EdgeUsagePattern, RecurrentEdgeStorageNeed]
        + EDGE_COMPONENT_CLASSES + EDGE_COMPUTER_COMPONENT_CLASSES + [EdgeDevice] + EDGE_DEVICE_BUILDER_CLASSES
        + SERVICE_CLASSES + SERVER_BUILDER_CLASSES
        + [Job, GPUJob] + [RecurrentEdgeDeviceNeed, RecurrentServerNeed] + RECURRENT_EDGE_COMPONENT_NEED_CLASSES
        + RECURRENT_EDGE_DEVICE_NEED_BUILDER_CLASSES + SERVICE_JOB_CLASSES
        + [Network] + SERVER_CLASSES + [Storage, System])

CANONICAL_COMPUTATION_ORDER = [
    UsageJourneyStep, UsageJourney, Device, Country, UsagePattern, EdgeUsageJourney, EdgeFunction,
    EdgeUsagePattern, RecurrentEdgeDeviceNeed, RecurrentServerNeed, RecurrentEdgeComponentNeed, EdgeComponent,
    EdgeDevice, Service, JobBase, Network, ServerBase, Storage, System]

ALL_CANONICAL_CLASSES_DICT = {cls.__name__: cls for cls in CANONICAL_COMPUTATION_ORDER}
ALL_CONCRETE_EFOOTPRINT_CLASSES_DICT = {cls.__name__: cls for cls in ALL_EFOOTPRINT_CLASSES}
ALL_EFOOTPRINT_CLASSES_DICT = copy(ALL_CONCRETE_EFOOTPRINT_CLASSES_DICT)
ALL_EFOOTPRINT_CLASSES_DICT.update(ALL_CANONICAL_CLASSES_DICT)



from typing import List, TYPE_CHECKING
from abc import abstractmethod

import numpy as np
from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.core.hardware.infra_hardware import InfraHardware
from efootprint.core.hardware.hardware_base import InsufficientCapacityError
from efootprint.abstract_modeling_classes.source_objects import SOURCE_VALUE_DEFAULT_NAME, SourceObject
from efootprint.constants.units import u
from efootprint.core.hardware.storage import Storage

if TYPE_CHECKING:
    from efootprint.core.usage.job import JobBase
    from efootprint.builders.services.service_base_class import Service


class ServerTypes:
    @classmethod
    def autoscaling(cls):
        return SourceObject("autoscaling")

    @classmethod
    def on_premise(cls):
        return SourceObject("on-premise")

    @classmethod
    def serverless(cls):
        return SourceObject("serverless")

    @classmethod
    def all(cls):
        return [cls.autoscaling(), cls.on_premise(), cls.serverless()]


class ServerBase(InfraHardware):
    @abstractmethod
    def _abc_marker(self):  # private abstract method so that this class is considered abstract
        pass

    default_values = {}

    list_values =  {"server_type": ServerTypes.all()}

    conditional_list_values =  {
            "fixed_nb_of_instances": {
                "depends_on": "server_type",
                "conditional_list_values": {
                    ServerTypes.autoscaling(): [EmptyExplainableObject()],
                    ServerTypes.serverless(): [EmptyExplainableObject()]
                }
            }
        }

    @classmethod
    def installable_services(cls) -> List:
        from efootprint.all_classes_in_order import SERVICE_CLASSES
        installable_services = []
        for service_class in SERVICE_CLASSES:
            for installable_on_class in service_class.installable_on():
                if issubclass(cls, installable_on_class):
                    installable_services.append(service_class)
                    break

        return installable_services


    def __init__(self, name: str, server_type: ExplainableObject, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity,
                 ram: ExplainableQuantity, compute: ExplainableQuantity,
                 power_usage_effectiveness: ExplainableQuantity, average_carbon_intensity: ExplainableQuantity,
                 utilization_rate: ExplainableQuantity, base_ram_consumption: ExplainableQuantity,
                 base_compute_consumption: ExplainableQuantity, storage: Storage,
                 fixed_nb_of_instances: ExplainableQuantity | EmptyExplainableObject = None):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan)
        self.hour_by_hour_compute_need = EmptyExplainableObject()
        self.hour_by_hour_ram_need = EmptyExplainableObject()
        self.available_compute_per_instance = EmptyExplainableObject()
        self.available_ram_per_instance = EmptyExplainableObject()
        self.raw_nb_of_instances = EmptyExplainableObject()
        self.nb_of_instances = EmptyExplainableObject()
        self.occupied_ram_per_instance = EmptyExplainableObject()
        self.occupied_compute_per_instance = EmptyExplainableObject()
        self.server_type = server_type.set_label(f"Server type of {self.name}")
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.ram = ram.set_label(f"RAM of {self.name}").to(u.GB_ram)
        self.compute = compute.set_label("tmp label")
        self.compute.set_label(f"Nb {self.compute_type.replace("_", " ")}s of {self.name}")
        self.power_usage_effectiveness = power_usage_effectiveness.set_label(f"PUE of {self.name}")
        self.average_carbon_intensity = average_carbon_intensity
        if SOURCE_VALUE_DEFAULT_NAME in self.average_carbon_intensity.label:
            self.average_carbon_intensity.set_label(f"Average carbon intensity of {self.name} electricity")
        self.utilization_rate = utilization_rate.set_label(f"{self.name} utilization rate")
        self.base_ram_consumption = base_ram_consumption.set_label(f"Base RAM consumption of {self.name}")
        self.base_compute_consumption = base_compute_consumption.set_label(
            f"Base {self.compute_type.replace("_", " ")} consumption of {self.name}")
        self.fixed_nb_of_instances = (fixed_nb_of_instances or EmptyExplainableObject()).set_label(
            f"User defined number of {self.name} instances").to(u.concurrent)
        self.storage = storage
        
    @property
    def compute_type(self) -> str:
        return str(self.compute.value.units)

    @property
    def calculated_attributes(self):
        return ["hour_by_hour_ram_need", "hour_by_hour_compute_need",
                "occupied_ram_per_instance", "occupied_compute_per_instance",
                "available_ram_per_instance", "available_compute_per_instance",
                "raw_nb_of_instances", "nb_of_instances",
                "instances_fabrication_footprint", "instances_energy", "energy_footprint"]

    @property
    def resources_unit_dict(self):
        return {"ram": "GB_ram", "compute": self.compute_type}

    @property
    def jobs(self) -> List["JobBase"]:
        from efootprint.core.usage.job import JobBase

        return (
            [modeling_obj for modeling_obj in self.modeling_obj_containers if isinstance(modeling_obj, JobBase)]
            + sum([service.jobs for service in self.installed_services], [])
            )


    @property
    def installed_services(self) -> List["Service"]:
        from efootprint.builders.services.service_base_class import Service

        return [modeling_obj for modeling_obj in self.modeling_obj_containers if isinstance(modeling_obj, Service)]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return [self.storage]

    def compute_hour_by_hour_resource_need(self, resource):
        resource_unit = u(self.resources_unit_dict[resource])
        hour_by_hour_resource_needs = EmptyExplainableObject()
        for job in self.jobs:
            hour_by_hour_resource_needs += (
                    job.hourly_avg_occurrences_across_usage_patterns * getattr(job, f"{resource}_needed"))

        return hour_by_hour_resource_needs.to(resource_unit).set_label(f"{self.name} hour by hour {resource} need")

    def update_hour_by_hour_ram_need(self):
        self.hour_by_hour_ram_need = self.compute_hour_by_hour_resource_need("ram")

    def update_hour_by_hour_compute_need(self):
        self.hour_by_hour_compute_need = self.compute_hour_by_hour_resource_need("compute")

    def update_occupied_ram_per_instance(self):
        self.occupied_ram_per_instance = (self.base_ram_consumption + sum(
            [service.base_ram_consumption for service in self.installed_services])).to(u.GB_ram).set_label(
            f"Occupied RAM per {self.name} instance including services")

    def update_occupied_compute_per_instance(self):
        self.occupied_compute_per_instance = (self.base_compute_consumption + sum(
            [service.base_compute_consumption for service in self.installed_services])).set_label(
            f"Occupied CPU per {self.name} instance including services")

    def update_available_ram_per_instance(self):
        available_ram_per_instance_before_services_installation = (self.ram * self.utilization_rate).to(u.GB_ram)
        available_ram_per_instance = (
                available_ram_per_instance_before_services_installation - self.occupied_ram_per_instance)
        if available_ram_per_instance.value < 0 * u.B_ram:
            raise InsufficientCapacityError(
                self, "RAM", available_ram_per_instance_before_services_installation, self.occupied_ram_per_instance)

        self.available_ram_per_instance = available_ram_per_instance.set_label(
            f"Available RAM per {self.name} instance")

    def update_available_compute_per_instance(self):
        available_compute_per_instance_before_services_installation = self.compute * self.utilization_rate
        available_compute_per_instance = (
                available_compute_per_instance_before_services_installation - self.occupied_compute_per_instance)
        if available_compute_per_instance.value < 0:
            raise InsufficientCapacityError(
                self, "compute", available_compute_per_instance_before_services_installation,
                self.occupied_compute_per_instance)

        self.available_compute_per_instance = available_compute_per_instance.set_label(
            f"Available CPU per {self.name} instance")

    def update_raw_nb_of_instances(self):
        nb_of_servers_based_on_ram_alone = (
                self.hour_by_hour_ram_need / self.available_ram_per_instance).to(u.concurrent).set_label(
            f"Raw nb of {self.name} instances based on RAM alone")
        nb_of_servers_based_on_cpu_alone = (
                self.hour_by_hour_compute_need / self.available_compute_per_instance).to(u.concurrent).set_label(
            f"Raw nb of {self.name} instances based on CPU alone")

        nb_of_servers_raw = nb_of_servers_based_on_ram_alone.np_compared_with(nb_of_servers_based_on_cpu_alone, "max")

        hour_by_hour_raw_nb_of_instances = nb_of_servers_raw.set_label(
            f"Hourly raw number of {self.name} instances")

        self.raw_nb_of_instances = hour_by_hour_raw_nb_of_instances

    def update_instances_energy(self):
        energy_spent_by_one_idle_instance_over_one_hour = (
                self.idle_power * self.power_usage_effectiveness * ExplainableQuantity(1 * u.hour, "one hour"))
        extra_energy_spent_by_one_fully_active_instance_over_one_hour = (
                (self.power - self.idle_power) * self.power_usage_effectiveness
                * ExplainableQuantity(1 * u.hour, "one hour"))

        server_energy = (
                energy_spent_by_one_idle_instance_over_one_hour * self.nb_of_instances
                + extra_energy_spent_by_one_fully_active_instance_over_one_hour * self.raw_nb_of_instances)

        self.instances_energy = server_energy.to(u.kWh).set_label(
            f"Hourly energy consumed by {self.name} instances")

    def autoscaling_update_nb_of_instances(self):
        hour_by_hour_nb_of_instances = self.raw_nb_of_instances.ceil()

        self.nb_of_instances = hour_by_hour_nb_of_instances.generate_explainable_object_with_logical_dependency(
            self.server_type).set_label(f"Hourly number of {self.name} instances")

    def serverless_update_nb_of_instances(self):
        hour_by_hour_nb_of_instances = self.raw_nb_of_instances.copy()

        self.nb_of_instances = hour_by_hour_nb_of_instances.generate_explainable_object_with_logical_dependency(
            self.server_type).set_label(f"Hourly number of {self.name} instances")

    def on_premise_update_nb_of_instances(self):
        if isinstance(self.raw_nb_of_instances, EmptyExplainableObject):
            nb_of_instances = EmptyExplainableObject(left_parent=self.raw_nb_of_instances)
        else:
            max_nb_of_instances = self.raw_nb_of_instances.max().ceil().to(u.concurrent)

            if not isinstance(self.fixed_nb_of_instances, EmptyExplainableObject):
                if max_nb_of_instances > self.fixed_nb_of_instances:
                    raise InsufficientCapacityError(
                        self, "number of instances", self.fixed_nb_of_instances, max_nb_of_instances)
                else:
                    fixed_nb_of_instances_np = Quantity(
                        np.full(len(self.raw_nb_of_instances), np.float32(self.fixed_nb_of_instances.magnitude)),
                        u.concurrent)
                    nb_of_instances = ExplainableHourlyQuantities(
                        fixed_nb_of_instances_np, self.raw_nb_of_instances.start_date, "Nb of instances",
                        left_parent=self.raw_nb_of_instances, right_parent=self.fixed_nb_of_instances)
            else:
                nb_of_instances_np = Quantity(
                    np.float32(max_nb_of_instances.magnitude) * np.ones(len(self.raw_nb_of_instances), dtype=np.float32),
                    u.concurrent)

                nb_of_instances = ExplainableHourlyQuantities(
                    nb_of_instances_np, self.raw_nb_of_instances.start_date,f"Hourly number of {self.name} instances",
                    left_parent=self.raw_nb_of_instances, right_parent=self.fixed_nb_of_instances,
                    operator="depending on not being empty")

        self.nb_of_instances = nb_of_instances.generate_explainable_object_with_logical_dependency(
            self.server_type).set_label(f"Hourly number of {self.name} instances")

    def update_nb_of_instances(self):
        logic_mapping = {
            ServerTypes.autoscaling(): self.autoscaling_update_nb_of_instances,
            ServerTypes.on_premise(): self.on_premise_update_nb_of_instances,
            ServerTypes.serverless(): self.serverless_update_nb_of_instances
        }
        logic_mapping[self.server_type]()



from copy import deepcopy
from time import perf_counter

from efootprint.abstract_modeling_classes.explainable_dict import ExplainableDict

start = perf_counter()
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceObject, SourceValue

from efootprint.builders.hardware.boaviztapi_utils import call_boaviztapi
from efootprint.constants.units import u
from efootprint.core.hardware.server import Server
from efootprint.core.hardware.server_base import ServerTypes
from efootprint.core.hardware.storage import Storage
from efootprint.logger import logger

boavizta_cloud_providers_request_result = call_boaviztapi(
        "https://api.boavizta.org/v1/cloud/instance/all_providers")
all_boavizta_cloud_providers = []
for cloud_provider in boavizta_cloud_providers_request_result:
    all_boavizta_cloud_providers.append(SourceObject(cloud_provider))
instance_types_conditional_list_values_dict = {"depends_on": "provider", "conditional_list_values": {}}
for cloud_provider in boavizta_cloud_providers_request_result:
    provider_instance_types = call_boaviztapi(
            f"https://api.boavizta.org/v1/cloud/instance/all_instances",
            params={"provider": cloud_provider}
    )
    instance_types_conditional_list_values_dict["conditional_list_values"][SourceObject(cloud_provider)] = [
        SourceObject(instance_type) for instance_type in provider_instance_types]


class BoaviztaCloudServer(Server):
    default_values = {
            "provider": SourceObject("scaleway"),
            "instance_type": SourceObject("ent1-s"),
            "server_type": ServerTypes.autoscaling(),
            "average_carbon_intensity": SourceValue(400 * u.g / u.kWh),
            "lifespan": SourceValue(6 * u.year),
            "idle_power": SourceValue(0 * u.W),
            "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless),
            "utilization_rate": SourceValue(0.9 * u.dimensionless),
            "base_ram_consumption": SourceValue(0 * u.GB_ram),
            "base_compute_consumption": SourceValue(0 * u.cpu_core),
            "fixed_nb_of_instances": EmptyExplainableObject()
        }


    server_list_values = deepcopy(Server.list_values)
    server_list_values.update({"provider": all_boavizta_cloud_providers})

    list_values = server_list_values

    server_conditional_list_values = deepcopy(Server.conditional_list_values)
    server_conditional_list_values.update({"instance_type": instance_types_conditional_list_values_dict})

    conditional_list_values = server_conditional_list_values

    def __init__(
            self, name: str, provider: ExplainableObject, instance_type: ExplainableObject, server_type: ExplainableObject,
            lifespan: ExplainableQuantity, idle_power: ExplainableQuantity,
            power_usage_effectiveness: ExplainableQuantity, average_carbon_intensity: ExplainableQuantity,
            utilization_rate: ExplainableQuantity, base_ram_consumption: ExplainableQuantity,
            base_compute_consumption: ExplainableQuantity, storage: Storage,
            fixed_nb_of_instances: ExplainableQuantity | EmptyExplainableObject | None = None):
        super().__init__(
            name, server_type, carbon_footprint_fabrication=SourceValue(0 * u.kg), power=SourceValue(0 * u.W),
            lifespan=lifespan, idle_power=idle_power, ram=SourceValue(0 * u.GB), compute=SourceValue(0 * u.cpu_core),
            power_usage_effectiveness=power_usage_effectiveness, average_carbon_intensity=average_carbon_intensity,
            utilization_rate=utilization_rate, base_ram_consumption=base_ram_consumption,
            base_compute_consumption=base_compute_consumption, storage=storage,
            fixed_nb_of_instances=fixed_nb_of_instances)
        self.provider = provider.set_label(f"{self.name} cloud provider")
        self.instance_type = instance_type.set_label(f"{self.name} instance type")
        self.impact_url = "https://api.boavizta.org/v1/cloud/instance"
        self.api_call_response = EmptyExplainableObject()

    @property
    def calculated_attributes(self):
        return (["api_call_response", "carbon_footprint_fabrication", "power", "ram", "compute"]
                + super().calculated_attributes)

    @property
    def attributes_that_shouldnt_trigger_update_logic(self):
        return super().attributes_that_shouldnt_trigger_update_logic + ["impact_url"]

    def update_api_call_response(self):
        params = {"provider": self.provider.value, "instance_type": self.instance_type.value}
        impact_source = Source(name="Boavizta API cloud instances",
                               link=f"{self.impact_url}?{'&'.join([key + '=' + params[key] for key in params])}")

        call_response = call_boaviztapi(url=self.impact_url, params=params)
        self.api_call_response = ExplainableDict(
            call_response, "API call response",
            left_parent=self.provider, right_parent=self.instance_type, operator="combined in Boavizta API call with",
            source=impact_source)

    def update_carbon_footprint_fabrication(self):
        self.carbon_footprint_fabrication = ExplainableQuantity(
            float(self.api_call_response.value["impacts"]["gwp"]["embedded"]["value"]) * u.kg,
            f"{self.name} fabrication carbon footprint", left_parent=self.api_call_response,
            operator="data extraction from", source=self.api_call_response.source)

    def update_power(self):
        average_power_unit = self.api_call_response.value["verbose"]["avg_power"]["unit"]
        use_time_ratio = self.api_call_response.value["verbose"]["use_time_ratio"]["value"]
        assert average_power_unit == "W", f"Unexpected power unit {average_power_unit}"
        assert float(use_time_ratio) == 1, f"Unexpected use time ratio {use_time_ratio}"
        average_power_value = float(self.api_call_response.value["verbose"]["avg_power"]["value"])

        self.power = ExplainableQuantity(
            average_power_value * u.W, f"{self.name} power", left_parent=self.api_call_response,
            operator="data extraction from", source=self.api_call_response.source)

    def update_ram(self):
        assert self.api_call_response.value["verbose"]["memory"]["unit"] == "GB", \
            f"Unexpected RAM unit {self.api_call_response.value['verbose']['memory']['unit']}"
        ram_spec = float(self.api_call_response.value["verbose"]["memory"]["value"])

        self.ram = ExplainableQuantity(
            ram_spec * u.GB_ram, f"{self.name} ram",
            left_parent=self.api_call_response, operator="data extraction from", source=self.api_call_response.source)

    def update_compute(self):
        nb_vcpu = float(self.api_call_response.value["verbose"]["vcpu"]["value"])

        self.compute = ExplainableQuantity(
            nb_vcpu * u.cpu_core, f"{self.name} compute",
            left_parent=self.api_call_response, operator="data extraction from", source=self.api_call_response.source)

logger.info(f"Imported BoaviztaCloudServer in {perf_counter() - start:.5f} seconds.")


if __name__ == "__main__":
    from efootprint.abstract_modeling_classes.explainable_object_base_class import \
        retrieve_update_function_from_mod_obj_and_attr_name

    for provider in all_boavizta_cloud_providers:
        for instance_type in instance_types_conditional_list_values_dict["conditional_list_values"][provider]:
            try:
                cloud_server = BoaviztaCloudServer(name=f"test_{provider}_{instance_type}",
                                    provider=SourceObject(provider.value), instance_type=SourceObject(instance_type.value),
                                    server_type=ServerTypes.autoscaling(),
                                    lifespan=SourceValue(6 * u.year),
                                    idle_power=SourceValue(0 * u.W),
                                    power_usage_effectiveness=SourceValue(1.2 * u.dimensionless),
                                    average_carbon_intensity=SourceValue(0.233 * u.kg / u.kWh),
                                    utilization_rate=SourceValue(0.9 * u.dimensionless),
                                    base_ram_consumption=SourceValue(0 * u.GB),
                                    base_compute_consumption=SourceValue(0 * u.cpu_core),
                                    storage=Storage.ssd())
                for attr_name in ["api_call_response", "carbon_footprint_fabrication", "power", "ram", "compute"]:
                    update_func = retrieve_update_function_from_mod_obj_and_attr_name(cloud_server, attr_name)
                    update_func()
                logger.info(f"{provider} - {instance_type}: Compute {cloud_server.compute} RAM {cloud_server.ram} "
                            f"CCF {cloud_server.carbon_footprint_fabrication} power {cloud_server.power}.")
            except Exception as e:
                logger.error(f"Error with provider {provider} and instance type {instance_type}: {e}")



from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source


@ExplainableObject.register_subclass(lambda d: "value" in d and "unit" not in d and isinstance(d["value"], dict))
class ExplainableDict(ExplainableObject):
    __slots__ = ()

    @classmethod
    def from_json_dict(cls, d):
        source = Source.from_json_dict(d["source"]) if "source" in d else None
        return cls(d["value"], label=d["label"], source=source)

    def to_json(self, save_calculated_attributes=False):
        output_dict = {"value": self.value}
        output_dict.update(super().to_json(save_calculated_attributes))

        return output_dict



from typing import List

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.edge.edge_device import EdgeDevice
from efootprint.core.hardware.edge.edge_ram_component import EdgeRAMComponent
from efootprint.core.hardware.edge.edge_cpu_component import EdgeCPUComponent
from efootprint.core.hardware.edge.edge_storage import EdgeStorage


class EdgeComputerRAMComponent(EdgeRAMComponent):
    def __init__(self, name: str):
        super().__init__(
            name=name,
            carbon_footprint_fabrication=SourceValue(0 * u.kg),
            power=SourceValue(0 * u.W),
            lifespan=SourceValue(1 * u.year),
            idle_power=SourceValue(0 * u.W),
            ram=SourceValue(1 * u.GB_ram),
            base_ram_consumption=SourceValue(0 * u.GB_ram))

    @property
    def calculated_attributes(self):
        return ["ram", "base_ram_consumption", "lifespan"] + super().calculated_attributes

    def update_ram(self):
        edge_device = self.edge_device
        if edge_device:
            self.ram = self.edge_device.ram.copy().set_label(f"RAM of {self.name}")
        else:
            self.ram = EmptyExplainableObject()

    def update_base_ram_consumption(self):
        edge_device = self.edge_device
        if edge_device:
            self.base_ram_consumption = self.edge_device.base_ram_consumption.copy().set_label(
                f"Base RAM consumption of {self.name}")
        else:
            self.base_ram_consumption = EmptyExplainableObject()

    def update_lifespan(self):
        edge_device = self.edge_device
        if edge_device:
            self.lifespan = self.edge_device.lifespan.copy().set_label(f"Lifespan of {self.name}")
        else:
            self.lifespan = EmptyExplainableObject()


class EdgeComputerCPUComponent(EdgeCPUComponent):
    def __init__(self, name: str):
        super().__init__(
            name=name,
            carbon_footprint_fabrication=SourceValue(0 * u.kg),
            power=SourceValue(1 * u.W),
            lifespan=SourceValue(1 * u.year),
            idle_power=SourceValue(0 * u.W),
            compute=SourceValue(1 * u.cpu_core),
            base_compute_consumption=SourceValue(0 * u.cpu_core))

    @property
    def calculated_attributes(self):
        return ["compute", "base_compute_consumption", "lifespan", "power", "idle_power"] + super().calculated_attributes

    def update_compute(self):
        edge_device = self.edge_device
        if edge_device:
            self.compute = edge_device.compute.copy().set_label(f"Compute of {self.name}")
        else:
            self.compute = EmptyExplainableObject()

    def update_base_compute_consumption(self):
        edge_device = self.edge_device
        if edge_device:
            self.base_compute_consumption = self.edge_device.base_compute_consumption.copy().set_label(
                f"Base compute consumption of {self.name}")
        else:
            self.base_compute_consumption = EmptyExplainableObject()

    def update_lifespan(self):
        edge_device = self.edge_device
        if edge_device:
            self.lifespan = self.edge_device.lifespan.copy().set_label(f"Lifespan of {self.name}")
        else:
            self.lifespan = EmptyExplainableObject()

    def update_power(self):
        edge_device = self.edge_device
        if edge_device:
            self.power = self.edge_device.power.copy().set_label(f"Power of {self.name}")
        else:
            self.power = EmptyExplainableObject()


    def update_idle_power(self):
        edge_device = self.edge_device
        if edge_device:
            self.idle_power = self.edge_device.idle_power.copy().set_label(f"Idle power of {self.name}")
        else:
            self.idle_power = EmptyExplainableObject()


class EdgeComputer(EdgeDevice):
    default_values = {
        "carbon_footprint_fabrication": SourceValue(60 * u.kg),
        "power": SourceValue(30 * u.W),
        "lifespan": SourceValue(6 * u.year),
        "idle_power": SourceValue(5 * u.W),
        "ram": SourceValue(8 * u.GB_ram),
        "compute": SourceValue(4 * u.cpu_core),
        "base_ram_consumption": SourceValue(1 * u.GB_ram),
        "base_compute_consumption": SourceValue(0.1 * u.cpu_core),
    }

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity,
                 ram: ExplainableQuantity, compute: ExplainableQuantity,
                 base_ram_consumption: ExplainableQuantity, base_compute_consumption: ExplainableQuantity,
                 storage: EdgeStorage):
        super().__init__(
            name=name,
            structure_carbon_footprint_fabrication=SourceValue(0 * u.kg),
            components=[],
            lifespan=lifespan)
        self.storage = storage
        self.carbon_footprint_fabrication = carbon_footprint_fabrication.set_label(
            f"Carbon footprint fabrication of {self.name}")
        self.power = power.set_label(f"Power of {self.name}")
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.ram = ram.set_label(f"RAM of {self.name}")
        self.compute = compute.set_label(f"Compute of {self.name}")
        self.base_ram_consumption = base_ram_consumption.set_label(f"Base RAM consumption of {self.name}")
        self.base_compute_consumption = base_compute_consumption.set_label(f"Base compute consumption of {self.name}")

    @property
    def calculated_attributes(self):
        return ["structure_carbon_footprint_fabrication"] + super().calculated_attributes

    @property
    def attribute_update_entanglements(self):
        return {"storage": self.generate_process_changes_from_storage_change}

    def generate_process_changes_from_storage_change(self, change: List[EdgeStorage]):
        old_storage, new_storage = change[0], change[1]
        component_needs_changes = [
            [edge_process.storage_need.edge_component, new_storage] for edge_process in self.recurrent_needs
        ]
        component_needs_changes.append([
            self.components, [self.cpu_component, self.ram_component, new_storage]
        ])

        return component_needs_changes

    def update_structure_carbon_footprint_fabrication(self):
        self.structure_carbon_footprint_fabrication = self.carbon_footprint_fabrication.copy().set_label(
            f"Structure fabrication carbon footprint of {self.name}")

    def after_init(self):
        if not self.components:
            ram_component = EdgeComputerRAMComponent(name=f"{self.name} RAM")
            cpu_component = EdgeComputerCPUComponent(name=f"{self.name} CPU")

            self.components = [cpu_component, ram_component, self.storage]
        super().after_init()

    @property
    def ram_component(self) -> EdgeComputerRAMComponent:
        return next(comp for comp in self.components if isinstance(comp, EdgeComputerRAMComponent))

    @property
    def cpu_component(self) -> EdgeComputerCPUComponent:
        return next(comp for comp in self.components if isinstance(comp, EdgeComputerCPUComponent))

    @property
    def available_ram_per_instance(self):
        return self.ram_component.available_ram_per_instance

    @property
    def available_compute_per_instance(self):
        return self.cpu_component.available_compute_per_instance

    @property
    def unitary_hourly_ram_need_per_usage_pattern(self):
        return self.ram_component.unitary_hourly_ram_need_per_usage_pattern

    @property
    def unitary_hourly_compute_need_per_usage_pattern(self):
        return self.cpu_component.unitary_hourly_compute_need_per_usage_pattern

    def self_delete(self):
        ram = self.ram_component
        cpu = self.cpu_component
        super().self_delete()
        for component in [ram, cpu]:
            component.self_delete()



from typing import List, Type, TYPE_CHECKING

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.core.hardware.server import Server
from efootprint.core.hardware.storage import Storage
from efootprint.core.usage.usage_journey_step import UsageJourneyStep
from efootprint.core.usage.job import Job

if TYPE_CHECKING:
    from efootprint.core.usage.usage_pattern import UsagePattern


class UsageJourney(ModelingObject):
    def __init__(self, name: str, uj_steps: List[UsageJourneyStep]):
        super().__init__(name)
        self.duration = EmptyExplainableObject()
        self.uj_steps = uj_steps

    def after_init(self):
        super().after_init()
        self.compute_calculated_attributes()

    @property
    def calculated_attributes(self):
        return ["duration"]

    @property
    def servers(self) -> List[Server]:
        servers = set()
        for job in self.jobs:
            servers = servers | {job.server}

        return list(servers)

    @property
    def storages(self) -> List[Storage]:
        storages = set()
        for job in self.jobs:
            storages = storages | {job.server.storage}

        return list(storages)

    @property
    def usage_patterns(self):
        return self.modeling_obj_containers

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List["UsagePattern"] | List[Job]:
        if self.usage_patterns:
            return self.usage_patterns
        else:
            return self.jobs

    @property
    def jobs(self) -> List[Job]:
        output_list = []
        for uj_step in self.uj_steps:
            output_list += uj_step.jobs

        return output_list

    def update_duration(self):
        user_time_spent_sum = sum(
            [uj_step.user_time_spent for uj_step in self.uj_steps], start=EmptyExplainableObject())

        self.duration = user_time_spent_sum.set_label(f"Duration of {self.name}")



import numbers
from copy import copy, deepcopy

import numpy as np
from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source
from efootprint.constants.units import get_unit


@ExplainableObject.register_subclass(lambda d: "value" in d and "unit" in d)
class ExplainableQuantity(ExplainableObject):
    __slots__ = (
        '_ExplainableHourlyQuantities',
        '_EmptyExplainableObject',
        'json_value_data',
    )

    @classmethod
    def from_json_dict(cls, d):
        value = {key: d[key] for key in ["value", "unit"]}
        source = Source.from_json_dict(d.get("source")) if d.get("source") else None
        return cls(value, label=d["label"], source=source)

    def __init__(
            self, value: Quantity | dict, label: str = None, left_parent: ExplainableObject = None,
            right_parent: ExplainableObject = None, operator: str = None, source: Source = None):
        from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
        from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
        self._ExplainableHourlyQuantities = ExplainableHourlyQuantities
        self._EmptyExplainableObject = EmptyExplainableObject
        self.json_value_data = None
        if isinstance(value, Quantity):
            if not isinstance(value.magnitude, float):
                value = float(value.magnitude) * value.units
            super().__init__(value, label, left_parent, right_parent, operator, source)
        elif isinstance(value, dict):
            self.json_value_data = value
            super().__init__(None, label, left_parent, right_parent, operator, source)
        else:
            raise ValueError(f"Variable 'value' of type {type(value)} isn’t a Quantity or a dict")

    @property
    def value(self):
        if self._value is None and self.json_value_data is not None:
            self._value = Quantity(
                float(self.json_value_data["value"]), get_unit(self.json_value_data["unit"]))

        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value
        self.json_value_data = None

    @value.deleter
    def value(self):
        self._value = None
        self.json_value_data = None

    @property
    def unit(self):
        return self.value.units

    def to(self, unit_to_convert_to):
        self.value = self.value.to(unit_to_convert_to)

        return self

    @property
    def magnitude(self):
        return self.value.magnitude

    def compare_with_and_return_max(self, other):
        if isinstance(other, ExplainableQuantity):
            if self.value >= other.value:
                return ExplainableQuantity(self.value, left_parent=self, right_parent=other, operator="max")
            else:
                return ExplainableQuantity(other.value, left_parent=self, right_parent=other, operator="max")
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def ceil(self):
        return ExplainableQuantity(np.ceil(self.value), left_parent=self, operator="ceil")

    def abs(self):
        return ExplainableQuantity(np.abs(self.value), left_parent=self, operator="abs")

    def copy(self):
        return ExplainableQuantity(copy(self.value), label=self.label, left_parent=self, operator="duplicate")

    def __gt__(self, other):
        if isinstance(other, ExplainableQuantity):
            return self.value > other.value
        elif isinstance(other, self._EmptyExplainableObject):
            return self.value > 0
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __lt__(self, other):
        if isinstance(other, ExplainableQuantity):
            return self.value < other.value
        elif isinstance(other, self._EmptyExplainableObject):
            return self.value < 0
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __eq__(self, other):
        if isinstance(other, ExplainableQuantity):
            converted_other_value = other.value.to(self.value.units)
            return np.allclose(
                np.asarray(self.magnitude), np.asarray(converted_other_value.magnitude), rtol=1e-4, atol=1e-6)
        elif isinstance(other, self._EmptyExplainableObject):
            return self.value == 0

        return False

    def __add__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            # summing with sum() adds an implicit 0 as starting value
            return ExplainableQuantity(self.value, left_parent=self, operator="")
        elif isinstance(other, self._EmptyExplainableObject):
            return ExplainableQuantity(self.value, left_parent=self, right_parent=other, operator="+")
        elif isinstance(other, ExplainableQuantity):
            return ExplainableQuantity(self.value + other.value, "", self, other, "+")
        elif isinstance(other, self._ExplainableHourlyQuantities):
            return other.__add__(self)
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableQuantity or ExplainableHourlyQuantities, "
                f"not with {type(other)}")

    def __sub__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            return ExplainableQuantity(self.value, left_parent=self, operator="")
        elif isinstance(other, self._EmptyExplainableObject):
            return ExplainableQuantity(self.value, left_parent=self, right_parent=other, operator="-")
        elif isinstance(other, ExplainableQuantity):
            return ExplainableQuantity(self.value - other.value, "", self, other, "-")
        elif isinstance(other, self._ExplainableHourlyQuantities):
            return other.__rsub__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __mul__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            return 0
        elif isinstance(other, self._EmptyExplainableObject):
            return self._EmptyExplainableObject(left_parent=self, right_parent=other, operator="*")
        elif isinstance(other, ExplainableQuantity):
            return ExplainableQuantity(self.value * other.value, "", self, other, "*")
        elif isinstance(other, self._ExplainableHourlyQuantities):
            return other.__mul__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __truediv__(self, other):
        if isinstance(other, ExplainableQuantity):
            return ExplainableQuantity(self.value / other.value, "", self, other, "/")
        elif isinstance(other, self._ExplainableHourlyQuantities):
            return other.__rtruediv__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        if isinstance(other, ExplainableQuantity):
            return ExplainableQuantity(other.value - self.value, "", other, self, "-")
        elif isinstance(other, self._ExplainableHourlyQuantities):
            return other.__sub__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity or ExplainableHourlyQuantities,"
                             f" not with {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rtruediv__(self, other):
        if isinstance(other, numbers.Number) and other == 0:
            return 0
        elif isinstance(other, self._EmptyExplainableObject):
            return self._EmptyExplainableObject(left_parent=other, right_parent=self, operator="/")
        elif isinstance(other, ExplainableQuantity):
            return ExplainableQuantity(other.value / self.value, "", other, self, "/")
        elif isinstance(other, self._ExplainableHourlyQuantities):
            return other.__truediv__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __round__(self, round_level):
        return ExplainableQuantity(
            round(self.value, round_level), label=self.label, left_parent=self,
            operator=f"rounded to {round_level} decimals", source=self.source)

    def to_json(self, save_calculated_attributes=False):
        if self.json_value_data is not None:
            output_dict = deepcopy(self.json_value_data)
        else:
            output_dict = {"value": float(self.value.magnitude), "unit": str(self.value.units)}

        output_dict.update(super().to_json(save_calculated_attributes))

        return output_dict

    def __repr__(self):
        return str(self)

    def __str__(self):
        if isinstance(self.value, Quantity):
            return f"{round(self.value, 2)}"
        else:
            return str(self.value)

    def __copy__(self):
        return ExplainableQuantity(copy(self.value), label=copy(self.label), source=copy(self.source))



from datetime import timedelta
from typing import Dict, List, Optional

from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.builders.services.service_base_class import Service
from efootprint.constants.units import u
from efootprint.builders.hardware.edge.edge_computer import EdgeComputer
from efootprint.core.country import Country
from efootprint.core.hardware.edge.edge_device import EdgeDevice
from efootprint.core.hardware.edge.edge_storage import EdgeStorage
from efootprint.core.hardware.network import Network
from efootprint.core.hardware.server import Server
from efootprint.core.hardware.storage import Storage
from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney
from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
from efootprint.core.usage.job import Job, JobBase
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.core.usage.usage_journey import UsageJourney
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.logger import logger
from efootprint.utils.tools import format_co2_amount, display_co2_amount


class System(ModelingObject):
    def __init__(self, name: str, usage_patterns: List[UsagePattern], edge_usage_patterns: List[EdgeUsagePattern],):
        super().__init__(name)
        self.total_footprint = EmptyExplainableObject()
        self.usage_patterns = usage_patterns
        self.edge_usage_patterns = edge_usage_patterns
        self.check_no_object_to_link_is_already_linked_to_another_system()
        self.simulation = None
        self.set_initial_and_previous_footprints()

    def set_initial_and_previous_footprints(self):
        self.previous_change = None
        self.previous_total_energy_footprints_sum_over_period = ExplainableObjectDict()
        self.previous_total_fabrication_footprints_sum_over_period = ExplainableObjectDict()
        self.all_changes = []
        self.initial_total_energy_footprints_sum_over_period = ExplainableObjectDict()
        self.initial_total_fabrication_footprints_sum_over_period = ExplainableObjectDict()

    def compute_calculated_attributes(self):
        self.check_no_object_to_link_is_already_linked_to_another_system()
        super().compute_calculated_attributes()

    @property
    def calculated_attributes(self) -> List[str]:
        return ["total_footprint"]

    @property
    def attributes_that_shouldnt_trigger_update_logic(self):
        return (super().attributes_that_shouldnt_trigger_update_logic
                + ["all_changes", "previous_change", "previous_total_energy_footprints_sum_over_period",
                   "previous_total_fabrication_footprints_sum_over_period",
                   "initial_total_energy_footprints_sum_over_period",
                   "initial_total_fabrication_footprints_sum_over_period", "simulation"])

    def check_no_object_to_link_is_already_linked_to_another_system(self):
        for mod_obj in self.all_linked_objects:
            mod_obj_systems = mod_obj.systems
            if mod_obj_systems and mod_obj_systems[0].id != self.id:
                raise PermissionError(f"{mod_obj.name} is already linked to {mod_obj_systems[0].name}, so it is "
                                      f"impossible to link it to {self.name}")
            if len(mod_obj_systems) > 1:
                raise ValueError(f"{mod_obj.name} is linked to 2 systems, this should never happen, please report an"
                                 f" e-footprint bug at https://github.com/Boavizta/e-footprint/issues")

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self):
        return self.edge_usage_patterns + self.usage_patterns

    @property
    def systems(self) -> List:
        return [self]

    def after_init(self):
        from time import perf_counter
        start = perf_counter()
        logger.info(f"Starting computing {self.name} modeling")
        mod_obj_computation_chain_excluding_self = self.mod_objs_computation_chain[1:]
        self.launch_mod_objs_computation_chain(mod_obj_computation_chain_excluding_self)
        self.compute_calculated_attributes()
        all_objects = self.all_linked_objects
        nb_of_calculated_attributes = sum([len(obj.calculated_attributes) for obj in all_objects])
        if nb_of_calculated_attributes > 0:
            compute_duration = round((perf_counter() - start), 3)
            logger.info(
                f"Computed {nb_of_calculated_attributes} calculated attributes over {len(all_objects)} objects in "
                f"{compute_duration} seconds or {round(1000 * compute_duration / nb_of_calculated_attributes, 2)} "
                f"ms per computation")
        self.initial_total_energy_footprints_sum_over_period = self.total_energy_footprint_sum_over_period
        self.initial_total_fabrication_footprints_sum_over_period = self.total_fabrication_footprint_sum_over_period
        self.trigger_modeling_updates = True

    def get_objects_linked_to_usage_patterns(
            self, usage_patterns: List[UsagePattern]) -> List[ModelingObject]:
        output_list =  self.storages + usage_patterns
        usage_journeys = self.usage_journeys
        uj_steps = list(set(sum([uj.uj_steps for uj in usage_journeys], start=[])))
        devices = list(set(sum([up.devices for up in usage_patterns], start=[])))
        all_modeling_objects = output_list + usage_journeys + uj_steps + devices

        return all_modeling_objects

    def get_objects_linked_to_edge_usage_patterns(
            self, edge_usage_patterns: List[EdgeUsagePattern]) -> List[ModelingObject]:
        output_list = self.edge_storages + edge_usage_patterns
        edge_usage_journeys = self.edge_usage_journeys
        edge_functions = list(set(sum([euj.edge_functions for euj in edge_usage_journeys], start=[])))
        recurrent_edge_device_needs = list(
            set(sum([ef.recurrent_edge_device_needs for ef in edge_functions], start=[])))
        recurrent_server_needs = list(
            set(sum([ef.recurrent_server_needs for ef in edge_functions], start=[])))
        recurrent_edge_component_needs = list(
            set(sum([redn.recurrent_edge_component_needs for redn in recurrent_edge_device_needs], start=[])))
        edge_devices = self.edge_devices
        edge_devices_components = list(set(sum([ed.components for ed in edge_devices], start=[])))
        all_modeling_objects = (
                output_list + edge_usage_journeys + edge_functions + recurrent_edge_device_needs
                + recurrent_server_needs + recurrent_edge_component_needs
                + edge_devices + edge_devices_components)

        return all_modeling_objects

    @property
    def all_linked_objects(self):
        return (self.networks + self.jobs + self.servers + self.services + self.countries
                + self.get_objects_linked_to_usage_patterns(self.usage_patterns)
                + self.get_objects_linked_to_edge_usage_patterns(self.edge_usage_patterns))

    @property
    def usage_journeys(self) -> List[UsageJourney]:
        return list(set([up.usage_journey for up in self.usage_patterns]))

    @property
    def edge_usage_journeys(self) -> List[EdgeUsageJourney]:
        return list(set([eup.edge_usage_journey for eup in self.edge_usage_patterns]))

    @property
    def countries(self) -> List[Country]:
        countries = list(set([up.country for up in self.usage_patterns]
                             + [eup.country for eup in self.edge_usage_patterns]))
        return countries

    @property
    def networks(self) -> List[Network]:
        return list(set([up.network for up in self.usage_patterns] + [eup.network for eup in self.edge_usage_patterns]))

    @property
    def jobs(self) -> List[JobBase]:
        jobs_from_usage_patterns = sum([up.jobs for up in self.usage_patterns], start=[])
        jobs_from_edge_usage_patterns = sum([eup.jobs for eup in self.edge_usage_patterns], start=[])
        return list(set(jobs_from_usage_patterns + jobs_from_edge_usage_patterns))

    @property
    def servers(self) -> List[Server]:
        return list(set([job.server for job in self.jobs if hasattr(job, "server")]))

    @property
    def services(self) -> List[Service]:
        return list(set(sum([server.installed_services for server in self.servers], start=[])))

    @property
    def edge_devices(self) -> List[EdgeDevice]:
        return list(set(sum([euj.edge_devices for euj in self.edge_usage_journeys], start=[])))

    @property
    def edge_computers(self) -> List[EdgeComputer]:
        return [hw for hw in self.edge_devices if isinstance(hw, EdgeComputer)]

    @property
    def storages(self) -> List[Storage]:
        return list(set([server.storage for server in self.servers]))

    @property
    def edge_storages(self) -> List[EdgeStorage]:
        edge_storages = []
        for edge_device in self.edge_devices:
            for component in edge_device.components:
                if isinstance(component, EdgeStorage):
                    edge_storages.append(component)
        return list(set(edge_storages))

    @staticmethod
    def get_efootprint_obj_by_name(
            efootprint_obj_name: str, efootprint_obj_list: List[ModelingObject]) -> Optional[ModelingObject]:
        for efootprint_obj in efootprint_obj_list:
            if efootprint_obj.name == efootprint_obj_name:
                return efootprint_obj
        return None

    @property
    def fabrication_footprints(self) -> Dict[str, Dict[str, ExplainableHourlyQuantities]]:
        fab_footprints = {
            "Servers": {server: server.instances_fabrication_footprint for server in self.servers},
            "Storage": {storage: storage.instances_fabrication_footprint for storage in self.storages},
            "Network": {},
            "Devices": {usage_pattern: usage_pattern.instances_fabrication_footprint
                        for usage_pattern in self.usage_patterns},
            "EdgeDevices": {edge_device: edge_device.instances_fabrication_footprint
                            for edge_device in self.edge_devices},
        }

        return fab_footprints

    @property
    def energy_footprints(self) -> Dict[str, Dict[str, ExplainableHourlyQuantities]]:
        energy_footprints = {
            "Servers": {server: server.energy_footprint for server in self.servers},
            "Storage": {storage: storage.energy_footprint for storage in self.storages},
            "Network": {network: network.energy_footprint for network in self.networks},
            "Devices": {usage_pattern: usage_pattern.energy_footprint
                        for usage_pattern in self.usage_patterns},
            "EdgeDevices": {edge_device: edge_device.energy_footprint
                            for edge_device in self.edge_devices},
        }

        return energy_footprints

    @property
    def total_fabrication_footprints(self) -> Dict[str, ExplainableHourlyQuantities]:
        fab_footprints = {
            "Servers": sum([server.instances_fabrication_footprint for server in self.servers],
                           start=EmptyExplainableObject()).to(u.kg).set_label(
                "Servers total fabrication footprint"),
            "Storage": sum([storage.instances_fabrication_footprint for storage in self.storages],
                           start=EmptyExplainableObject()).to(u.kg).set_label(
                "Storage total fabrication footprint"),
            "Network": EmptyExplainableObject(),
            "Devices": sum([usage_pattern.instances_fabrication_footprint
                           for usage_pattern in self.usage_patterns], start=EmptyExplainableObject()).to(u.kg).set_label(
                "Devices total fabrication footprint"),
            "EdgeDevices": sum([edge_device.instances_fabrication_footprint for edge_device in self.edge_devices],
                               start=EmptyExplainableObject()).to(u.kg).set_label(
                "Edge devices total fabrication footprint"),
        }

        return fab_footprints

    @property
    def total_energy_footprints(self) -> Dict[str, ExplainableHourlyQuantities]:
        energy_footprints = {
            "Servers": sum([server.energy_footprint for server in self.servers], start=EmptyExplainableObject()
                           ).to(u.kg).set_label("Servers total energy footprint"),
            "Storage": sum([storage.energy_footprint for storage in self.storages], start=EmptyExplainableObject()
                           ).to(u.kg).set_label("Storage total energy footprint"),
            "Network": sum([network.energy_footprint for network in self.networks], start=EmptyExplainableObject()
                           ).to(u.kg).set_label("Network total energy footprint"),
            "Devices": sum([usage_pattern.energy_footprint for usage_pattern in self.usage_patterns],
                           start=EmptyExplainableObject()).to(u.kg).set_label("Devices total energy footprint"),
            "EdgeDevices": sum([edge_device.energy_footprint for edge_device in self.edge_devices],
                               start=EmptyExplainableObject()).to(u.kg).set_label("Edge devices total energy footprint"),
        }

        return energy_footprints

    @staticmethod
    def sum_and_remove_empty_explainable_object(expl_obj):
        tmp_sum = expl_obj.sum()
        if isinstance(tmp_sum, EmptyExplainableObject):
            tmp_sum = ExplainableQuantity(0 * u.kg, "null value")

        return tmp_sum

    @property
    def fabrication_footprint_sum_over_period(self) -> Dict[str, Dict[ModelingObject, ExplainableQuantity]]:
        fab_footprints_sum = {}
        for category_key, category_dict in self.fabrication_footprints.items():
            fab_footprints_sum[category_key] = {
                obj_key: self.sum_and_remove_empty_explainable_object(obj_value).to(u.kg).set_label(
                    f"{obj_key} fabrication footprints summed over modeling period")
                for obj_key, obj_value in category_dict.items()
            }

        return fab_footprints_sum

    @property
    def energy_footprint_sum_over_period(self) -> Dict[str, Dict[ModelingObject, ExplainableQuantity]]:
        energy_footprints_sum = {}
        for key, dict_value in self.energy_footprints.items():
            energy_footprints_sum[key] = {
                obj_key: self.sum_and_remove_empty_explainable_object(obj_value).to(u.kg).set_label(
                    f"{obj_key} energy footprints summed over modeling period")
                for obj_key, obj_value in dict_value.items()
            }

        return energy_footprints_sum

    @property
    def total_fabrication_footprint_sum_over_period(self) -> Dict[str, ExplainableQuantity]:
        fab_footprints = {
            object_category: self.sum_and_remove_empty_explainable_object(category_value).to(u.kg).set_label(
                f"{object_category} total fabrication footprints summed over modeling period")
            for object_category, category_value in self.total_fabrication_footprints.items()
        }

        return ExplainableObjectDict(fab_footprints)

    @property
    def total_energy_footprint_sum_over_period(self) -> Dict[str, ExplainableQuantity]:
        energy_footprints = {
            object_category: self.sum_and_remove_empty_explainable_object(category_value).to(u.kg).set_label(
                f"{object_category} total energy footprints summed over modeling period")
            for object_category, category_value in self.total_energy_footprints.items()
        }

        return ExplainableObjectDict(energy_footprints)

    def update_total_footprint(self):
        total_footprint = (
            sum(
                [sum(self.fabrication_footprints[key].values()) + sum(self.energy_footprints[key].values())
                for key in self.fabrication_footprints], start=EmptyExplainableObject()
            )
        ).to(u.kg).set_label(f"{self.name} total carbon footprint")

        self.total_footprint = round(total_footprint, 4)

    def plot_footprints_by_category_and_object(self, filename=None, height=400, width=800, notebook=True):
        import plotly.express as px
        import plotly

        fab_footprints = self.fabrication_footprint_sum_over_period
        energy_footprints = self.energy_footprint_sum_over_period

        rows_as_dicts = []
        value_colname = "tonnes CO2 emissions"

        for category in fab_footprints:
            fab_objects = sorted(fab_footprints[category].items(), key=lambda x: x[0].name)
            energy_objects = sorted(energy_footprints[category].items(), key=lambda x: x[0].name)

            for objs, color in zip([energy_objects, fab_objects], ["Electricity", "Fabrication"]):
                for object, quantity in objs:
                    magnitude_kg = quantity.magnitude
                    magnitude_tonnes = magnitude_kg / 1000
                    amount_str = display_co2_amount(format_co2_amount(magnitude_kg))

                    rows_as_dicts.append({
                        "Type": color, "Category": category, "Object": object.name, value_colname: magnitude_tonnes,
                        "Amount": amount_str})

        import pandas as pd
        df = pd.DataFrame.from_records(rows_as_dicts)

        total_co2 = df[value_colname].sum()
        total_footprint = self.total_footprint

        start_date = total_footprint.start_date
        end_date = start_date + timedelta(hours=len(total_footprint.value) - 1)

        fig = px.bar(
            df, x="Category", y=value_colname, color='Type', barmode='group',
            height=height, width=width,
            hover_data={"Type": False, "Category": False, "Object": True, value_colname: False, "Amount": True},
            template="plotly_white",
            title=f"Total CO2 emissions from {start_date.date()} to {end_date.date()}: "
                  f"{display_co2_amount(format_co2_amount(total_co2 * 1000, rounding_value=0))}"
        )

        # Legend placement logic
        total_energy_servers = sum(energy_footprints["Servers"].values(), start=0)
        total_fab_servers = sum(fab_footprints["Servers"].values(), start=0)
        total_energy_devices = sum(energy_footprints["Devices"].values(), start=0)
        total_fab_devices = sum(fab_footprints["Devices"].values(), start=0)

        if (total_energy_servers + total_fab_servers) > (total_energy_devices + total_fab_devices):
            legend_alignment = "right"
            legend_x = 0.98
        else:
            legend_alignment = "left"
            legend_x = 0.02

        fig.update_layout(
            legend={"orientation": "v", "yanchor": "top", "y": 1.02, "xanchor": legend_alignment, "x": legend_x,
                    "title": ""},
            title={"x": 0.5, "y": 0.9, "xanchor": 'center', "yanchor": 'top'}
        )

        # Add annotations (percentages per category and type)
        total_by_cat_type = df.groupby(["Category", "Type"])[value_colname].sum()

        for (category, source_type), height_val in total_by_cat_type.items():
            x_shift = 30 if source_type == 'Fabrication' else -30
            percentage = int((height_val / total_co2) * 100)

            fig.add_annotation(
                x=category, y=height_val,
                text=f"{percentage}%",
                showarrow=False,
                yshift=10,
                xshift=x_shift
            )

        if notebook:
            from IPython.display import HTML

            if filename is None:
                filename = f"{self.name} footprints.html"

            plotly.offline.plot(fig, filename=filename, auto_open=False)

            return HTML(filename)
        else:
            return fig

    def plot_emission_diffs(self, filepath=None, figsize=(10, 5), from_start=False, plt_show=False):
        from matplotlib import pyplot as plt

        if self.previous_change is None:
            raise ValueError(
                f"There has been no change to the system yet so no diff to plot.\n"
                f"Use System.plot_footprints_by_category_and_object() to visualize footprints")

        if from_start and len(self.all_changes) > 1:
            emissions_dict__old = [self.initial_total_energy_footprints_sum_over_period,
                                   self.initial_total_fabrication_footprints_sum_over_period]
        else:
            emissions_dict__old = [self.previous_total_energy_footprints_sum_over_period,
                                   self.previous_total_fabrication_footprints_sum_over_period]

        emissions_dict__new = [self.total_energy_footprint_sum_over_period,
                               self.total_fabrication_footprint_sum_over_period]

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
        from efootprint.utils.plot_emission_diffs import EmissionPlotter
        EmissionPlotter(
            ax, emissions_dict__old, emissions_dict__new, rounding_value=0).plot_emission_diffs()

        if filepath is not None:
            plt.savefig(filepath, bbox_inches='tight')

        if plt_show:
            plt.show()

        return fig, ax



from typing import List

from efootprint.core.country import Country
from efootprint.constants.units import u
from efootprint.core.hardware.device import Device
from efootprint.core.usage.usage_journey import UsageJourney
from efootprint.core.usage.compute_nb_occurrences_in_parallel import compute_nb_avg_hourly_occurrences
from efootprint.core.usage.job import Job
from efootprint.core.hardware.network import Network
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import (
    ExplainableHourlyQuantities)
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject


class UsagePattern(ModelingObject):
    def __init__(self, name: str, usage_journey: UsageJourney, devices: List[Device],
                 network: Network, country: Country, hourly_usage_journey_starts: ExplainableHourlyQuantities):
        super().__init__(name)
        self.utc_hourly_usage_journey_starts = EmptyExplainableObject()
        self.nb_usage_journeys_in_parallel = EmptyExplainableObject()
        self.devices_energy = EmptyExplainableObject()
        self.devices_energy_footprint = EmptyExplainableObject()
        self.devices_fabrication_footprint = EmptyExplainableObject()
        self.energy_footprint = EmptyExplainableObject()
        self.instances_fabrication_footprint = EmptyExplainableObject()
        self.hourly_usage_journey_starts = hourly_usage_journey_starts.to(u.occurrence).set_label(
            f"{self.name} hourly nb of visits")
        self.usage_journey = usage_journey
        self.devices = devices
        self.network = network
        self.country = country

    @property
    def calculated_attributes(self):
        return ["utc_hourly_usage_journey_starts", "nb_usage_journeys_in_parallel", "devices_energy",
                "devices_energy_footprint", "devices_fabrication_footprint", "energy_footprint",
                "instances_fabrication_footprint"]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return self.jobs

    @property
    def jobs(self) -> List[Job]:
        return self.usage_journey.jobs

    def update_utc_hourly_usage_journey_starts(self):
        utc_hourly_usage_journey_starts = self.hourly_usage_journey_starts.convert_to_utc(
            local_timezone=self.country.timezone)

        self.utc_hourly_usage_journey_starts = utc_hourly_usage_journey_starts.set_label(f"{self.name} UTC")

    def update_nb_usage_journeys_in_parallel(self):
        nb_of_usage_journeys_in_parallel = compute_nb_avg_hourly_occurrences(
            self.utc_hourly_usage_journey_starts, self.usage_journey.duration)

        self.nb_usage_journeys_in_parallel = nb_of_usage_journeys_in_parallel.to(u.concurrent).set_label(
            f"{self.name} hourly nb of user journeys in parallel")

    def update_devices_energy(self):
        total_devices_energy_spent_over_one_full_hour = sum(
            [device.power for device in self.devices]) * ExplainableQuantity(1 * u.hour, "one full hour")

        devices_energy = (self.nb_usage_journeys_in_parallel * total_devices_energy_spent_over_one_full_hour).to(u.kWh)

        self.devices_energy = devices_energy.set_label(f"Energy consumed by {self.name} devices")

    def update_devices_energy_footprint(self):
        energy_footprint = (self.devices_energy * self.country.average_carbon_intensity).to(u.kg)
        
        self.devices_energy_footprint = energy_footprint.set_label(f"Devices energy footprint of {self.name}")

    def update_devices_fabrication_footprint(self):
        devices_fabrication_footprint_over_one_hour = EmptyExplainableObject()
        for device in self.devices:
            device_uj_fabrication_footprint = (
                    device.carbon_footprint_fabrication * ExplainableQuantity(1 * u.hour, "one hour")
                    / (device.lifespan * device.fraction_of_usage_time)
            ).to(u.g).set_label(
                f"{device.name} fabrication footprint over one hour")
            devices_fabrication_footprint_over_one_hour += device_uj_fabrication_footprint

        devices_fabrication_footprint = (
                self.nb_usage_journeys_in_parallel * devices_fabrication_footprint_over_one_hour).to(u.kg)

        self.devices_fabrication_footprint = devices_fabrication_footprint.set_label(
            f"Devices fabrication footprint of {self.name}")

    def update_energy_footprint(self):
        # The copy is to create a new ExplainableObject with the same value as the previous one, and hence make clear
        # in the calculation graph that the usage pattern energy footprint is the same as the devices energy footprint
        self.energy_footprint = self.devices_energy_footprint.copy().set_label(f"{self.name} total energy footprint")

    def update_instances_fabrication_footprint(self):
        self.instances_fabrication_footprint = self.devices_fabrication_footprint.copy().set_label(
            f"{self.name} total fabrication footprint")



from copy import copy
from typing import List, TYPE_CHECKING

import numpy as np
from pint import Quantity

from efootprint.constants.sources import Sources
from efootprint.core.hardware.edge.edge_component import EdgeComponent
from efootprint.core.hardware.hardware_base import InsufficientCapacityError
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern


class NegativeCumulativeStorageNeedError(Exception):
    def __init__(self, storage_obj: "EdgeStorage", cumulative_quantity: Quantity):
        self.storage_obj = storage_obj
        self.cumulative_quantity = cumulative_quantity

        message = (
            f"In EdgeStorage object {self.storage_obj.name}, negative cumulative storage need detected: "
            f"{np.min(cumulative_quantity):~P}. Please check your processes "
            f"or increase the base_storage_need value, currently set to {self.storage_obj.base_storage_need.value}")
        super().__init__(message)


class EdgeStorage(EdgeComponent):
    compatible_root_units = [u.bit]
    default_values = {
        "carbon_footprint_fabrication_per_storage_capacity": SourceValue(160 * u.kg / u.TB),
        "power_per_storage_capacity": SourceValue(1.3 * u.W / u.TB),
        "lifespan": SourceValue(6 * u.years),
        "idle_power": SourceValue(0 * u.W),
        "storage_capacity": SourceValue(1 * u.TB),
        "base_storage_need": SourceValue(30 * u.GB),
    }

    @classmethod
    def ssd(cls, name="Default SSD storage", **kwargs):
        output_args = {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(
                160 * u.kg / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "power_per_storage_capacity": SourceValue(1.3 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "lifespan": SourceValue(6 * u.years),
            "idle_power": SourceValue(0 * u.W),
            "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "base_storage_need": SourceValue(0 * u.TB),
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def hdd(cls, name="Default HDD storage", **kwargs):
        output_args = {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(
                20 * u.kg / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "power_per_storage_capacity": SourceValue(4.2 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "lifespan": SourceValue(4 * u.years),
            "idle_power": SourceValue(0 * u.W),
            "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "base_storage_need": SourceValue(0 * u.TB),
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def archetypes(cls):
        return [cls.ssd, cls.hdd]

    def __init__(self, name: str, storage_capacity: ExplainableQuantity,
                 carbon_footprint_fabrication_per_storage_capacity: ExplainableQuantity,
                 power_per_storage_capacity: ExplainableQuantity, idle_power: ExplainableQuantity,
                 base_storage_need: ExplainableQuantity, lifespan: ExplainableQuantity):
        super().__init__(
            name, carbon_footprint_fabrication=SourceValue(0 * u.kg), power=SourceValue(0 * u.W),
            lifespan=lifespan, idle_power=idle_power)
        self.carbon_footprint_fabrication_per_storage_capacity = (carbon_footprint_fabrication_per_storage_capacity
        .set_label(f"Fabrication carbon footprint of {self.name} per storage capacity"))
        self.power_per_storage_capacity = power_per_storage_capacity.set_label(
            f"Power of {self.name} per storage capacity")
        self.storage_capacity = storage_capacity.set_label(f"Storage capacity of {self.name}")
        self.base_storage_need = base_storage_need.set_label(f"{self.name} initial storage need")

        self.unitary_storage_delta_per_usage_pattern = ExplainableObjectDict()
        self.cumulative_unitary_storage_need_per_usage_pattern = ExplainableObjectDict()

    @property
    def calculated_attributes(self):
        return (["carbon_footprint_fabrication", "power", "unitary_storage_delta_per_usage_pattern",
                 "cumulative_unitary_storage_need_per_usage_pattern"] + super().calculated_attributes)

    def update_carbon_footprint_fabrication(self):
        self.carbon_footprint_fabrication = (
                self.carbon_footprint_fabrication_per_storage_capacity * self.storage_capacity).set_label(
            f"Carbon footprint of {self.name}")

    def update_power(self):
        self.power = (self.power_per_storage_capacity * self.storage_capacity).set_label(f"Power of {self.name}")

    def update_dict_element_in_unitary_storage_delta_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_storage_delta = sum(
            [need.unitary_hourly_need_per_usage_pattern[usage_pattern]
             for need in self.recurrent_edge_component_needs if usage_pattern in need.edge_usage_patterns],
            start=EmptyExplainableObject())

        self.unitary_storage_delta_per_usage_pattern[usage_pattern] = unitary_storage_delta.set_label(
            f"Hourly storage delta for {self.name} in {usage_pattern.name}")

    def update_unitary_storage_delta_per_usage_pattern(self):
        self.unitary_storage_delta_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_storage_delta_per_usage_pattern(usage_pattern)

    def update_dict_element_in_cumulative_unitary_storage_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_storage_delta = self.unitary_storage_delta_per_usage_pattern[usage_pattern]
        
        if isinstance(unitary_storage_delta, EmptyExplainableObject):
            self.cumulative_unitary_storage_need_per_usage_pattern[usage_pattern] = EmptyExplainableObject(
                left_parent=unitary_storage_delta)
        else:
            edge_computer_usage_span_in_hours = int(copy(
                usage_pattern.edge_usage_journey.usage_span.value).to(u.hour).magnitude)
            unitary_storage_delta_over_single_usage_span = ExplainableHourlyQuantities(
                Quantity(
                    unitary_storage_delta.magnitude[:edge_computer_usage_span_in_hours],
                    unitary_storage_delta.unit
                ),
                start_date=unitary_storage_delta.start_date,
                left_parent=unitary_storage_delta,
                right_parent=usage_pattern.edge_usage_journey.usage_span,
                operator="truncated by"
            )
            delta_array = np.copy(unitary_storage_delta_over_single_usage_span.value.magnitude)
            delta_unit = unitary_storage_delta_over_single_usage_span.value.units

            # Add base storage need to first hour
            delta_array[0] += self.base_storage_need.value.to(delta_unit).magnitude

            # Compute cumulative storage
            cumulative_array = np.cumsum(delta_array, dtype=np.float32)
            cumulative_quantity = Quantity(cumulative_array, delta_unit)

            if np.min(cumulative_quantity.magnitude) < 0:
                raise NegativeCumulativeStorageNeedError(self, cumulative_quantity)
            
            if np.max(cumulative_quantity) > self.storage_capacity.value:
                raise InsufficientCapacityError(
                    self, "storage capacity", self.storage_capacity,
                    ExplainableQuantity(cumulative_quantity.max(), label=f"{self.name} cumulative storage need for {usage_pattern.name}"))

            self.cumulative_unitary_storage_need_per_usage_pattern[usage_pattern] = ExplainableHourlyQuantities(
                cumulative_quantity,
                start_date=unitary_storage_delta_over_single_usage_span.start_date,
                label=f"Cumulative storage need for {self.name} in {usage_pattern.name}",
                left_parent=unitary_storage_delta_over_single_usage_span,
                right_parent=self.base_storage_need,
                operator="cumulative sum of storage delta with initial storage need"
            ).generate_explainable_object_with_logical_dependency(self.storage_capacity)

    def update_cumulative_unitary_storage_need_per_usage_pattern(self):
        self.cumulative_unitary_storage_need_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_cumulative_unitary_storage_need_per_usage_pattern(usage_pattern)

    def update_dict_element_in_unitary_power_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_storage_delta = self.unitary_storage_delta_per_usage_pattern[usage_pattern]

        if isinstance(unitary_storage_delta, EmptyExplainableObject):
            unitary_power = self.idle_power
        else:
            unitary_activity_level = (unitary_storage_delta.abs() / self.storage_capacity).to(u.dimensionless)
            unitary_power = self.idle_power + (self.power - self.idle_power) * unitary_activity_level

        self.unitary_power_per_usage_pattern[usage_pattern] = unitary_power.set_label(
            f"Hourly power for {self.name} in {usage_pattern.name}")

    def update_unitary_power_per_usage_pattern(self):
        self.unitary_power_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_power_per_usage_pattern(usage_pattern)



import json
import os
import time

from selenium import webdriver

from efootprint.utils.graph_tools import WIDTH, HEIGHT


def capture_screenshot(html_file_path, output_image_path, width, height):
    options = webdriver.ChromeOptions()
    options.headless = True
    driver = webdriver.Chrome(options=options)
    driver.set_window_size(width, height)

    driver.get(f"file://{html_file_path}")
    time.sleep(2)

    driver.save_screenshot(output_image_path)

    driver.quit()


def save_graph_as_both_html_and_png(input_graph, output_filepath, width=WIDTH, height=HEIGHT):
    input_graph.set_options(json.dumps({'physics': {"springLength": 250}}))
    input_graph.show(f"{output_filepath}")
    capture_screenshot(
        os.path.abspath(f"{output_filepath}"), f"{output_filepath.replace('html', 'png')}", width, height)
